# Copyright 2014-2020 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Tests for the factory functions to create proximal operators."""

from __future__ import division, print_function

import numpy as np
import pytest
import scipy.special

import odl
from odl.functionals.functional import FunctionalDefaultConvexConjugate
from odl.core.util.testutils import all_almost_equal, noise_element, simple_fixture

# --- pytest fixtures --- #


# Global pytest mark
pytestmark = pytest.mark.suite('largescale')


stepsize = simple_fixture('stepsize', [0.1, 1.0, 10.0])
linear_offset = simple_fixture('linear_offset', [False, True])
quadratic_offset = simple_fixture('quadratic_offset', [False, True])
dual = simple_fixture('dual', [False, True])


func_params = ['l1', 'l2', 'l2^2', 'kl', 'kl_cross_ent', 'const',
               'groupl1-1', 'groupl1-2',
               'nuclearnorm-1-1', 'nuclearnorm-1-2', 'nuclearnorm-1-inf',
               'quadratic', 'linear', 'huber']

func_ids = [" functional='{}' ".format(p) for p in func_params]


@pytest.fixture(scope="module", ids=func_ids, params=func_params)
def functional(request, linear_offset, quadratic_offset, dual, odl_impl_device_pairs):
    """Return functional whose proximal should be tested."""
    name = request.param.strip()
    impl, device = odl_impl_device_pairs
    space = odl.uniform_discr(0, 1, 2, impl=impl, device=device)

    if name == 'l1':
        func = odl.functionals.L1Norm(space)
    elif name == 'l2':
        func = odl.functionals.L2Norm(space)
    elif name == 'l2^2':
        func = odl.functionals.L2NormSquared(space)
    elif name == 'kl':
        func = odl.functionals.KullbackLeibler(space)
    elif name == 'kl_cross_ent':
        func = odl.functionals.KullbackLeiblerCrossEntropy(space)
    elif name == 'const':
        func = odl.functionals.ConstantFunctional(space, constant=2)
    elif name.startswith('groupl1'):
        exponent = float(name.split('-')[1])
        space = odl.ProductSpace(space, 2)
        func = odl.functionals.GroupL1Norm(space, exponent=exponent)
    elif name.startswith('nuclearnorm'):
        outer_exp = float(name.split('-')[1])
        singular_vector_exp = float(name.split('-')[2])

        space = odl.ProductSpace(odl.ProductSpace(space, 2), 3)
        func = odl.functionals.NuclearNorm(space,
                                       outer_exp=outer_exp,
                                       singular_vector_exp=singular_vector_exp)
    elif name == 'quadratic':
        func = odl.functionals.QuadraticForm(
            operator=odl.IdentityOperator(space),
            vector=space.one(),
            constant=0.623,
        )
    elif name == 'linear':
        func = odl.functionals.QuadraticForm(vector=space.one(), constant=0.623)
    elif name == 'huber':
        func = odl.functionals.Huber(space, gamma=0.162)
    else:
        assert False

    if quadratic_offset:
        if linear_offset:
            g = noise_element(space)
            if name.startswith('kl'):
                g = odl.abs(g)
        else:
            g = None

        quadratic_coeff = 1.32
        func = odl.functionals.FunctionalQuadraticPerturb(
            func, quadratic_coeff=quadratic_coeff, linear_term=g
        )

    elif linear_offset:
        g = noise_element(space)
        if name.startswith('kl'):
            g = odl.abs(g)
        func = func.translated(g)

    if dual:
        func = func.convex_conj

    return func


# Margin of error
EPS = 1e-6


# --- Functional tests --- #


def proximal_objective(functional, x, y):
    """Objective function of the proximal optimization problem."""
    return functional(y) + (1.0 / 2.0) * (x - y).norm() ** 2


def test_proximal_definition(functional, stepsize):
    """Test the definition of the proximal:

        prox[f](x) = argmin_y {f(y) + 1/2 ||x-y||^2}

    Hence we expect for all x in the domain of the proximal

        x* = prox[f](x)

        f(x*) + 1/2 ||x-x*||^2 <= f(y) + 1/2 ||x-y||^2
    """
    if isinstance(functional, FunctionalDefaultConvexConjugate):
        pytest.skip('functional has no call method')
        return

    # No implementation of the proximal for convex conj of
    # FunctionalQuadraticPerturb unless the quadratic term is 0.
    if (
        isinstance(functional, odl.functionals.FunctionalQuadraticPerturb)
        and functional.quadratic_coeff != 0
    ):
        pytest.skip('functional has no proximal')
        return

    # No implementation of the proximal for quardartic form
    if isinstance(functional, odl.functionals.QuadraticForm):
        pytest.skip('functional has no proximal')
        return

    # No implementation of the proximal for translations of quardartic form
    if (
        isinstance(functional, odl.functionals.FunctionalTranslation)
        and isinstance(functional.functional, odl.functionals.QuadraticForm)
    ):
        pytest.skip('functional has no proximal')
        return

    # No implementation of the proximal for convex conj of quardartic form,
    # except if the quadratic part is 0.
    if (
        isinstance(functional, odl.functionals.FunctionalQuadraticPerturb)
        and isinstance(functional.functional, odl.functionals.QuadraticForm)
        and functional.functional.operator is not None
    ):
        pytest.skip('functional has no proximal')
        return

    proximal = functional.proximal(stepsize)

    assert proximal.domain == functional.domain
    assert proximal.range == functional.domain

    for _ in range(100):
        x = noise_element(proximal.domain) * 10
        prox_x = proximal(x)
        f_prox_x = proximal_objective(stepsize * functional, x, prox_x)

        y = noise_element(proximal.domain)
        f_y = proximal_objective(stepsize * functional, x, y)

        if not f_prox_x <= f_y + EPS:
            print(repr(functional), x, y, prox_x, f_prox_x, f_y)

        assert f_prox_x <= f_y + EPS


def convex_conj_objective(functional, x, y):
    """Objective function of the convex conjugate problem."""
    return x.inner(y) - functional(x)


def func_convex_conj_has_call(functional):
    """Helper to determine whether a convex conjugate can be called."""
    f_cconj = functional.convex_conj
    if isinstance(f_cconj, FunctionalDefaultConvexConjugate):
        return False

    elif (
        isinstance(f_cconj, odl.functionals.FunctionalTranslation)
        and isinstance(f_cconj.functional, FunctionalDefaultConvexConjugate)
    ):
        return False

    return True


def test_convex_conj_defintion(functional):
    """Test the defintion of the convex conjugate:

        f^*(y) = sup_x {<x, y> - f(x)}

    Hence we expect for all x in the domain of the proximal

        <x, y> - f(x) <= f^*(y)
    """
    if isinstance(functional, FunctionalDefaultConvexConjugate):
        pytest.skip('`functional` has no `_call` implementation')
        return

    if not func_convex_conj_has_call(functional):
        pytest.skip('`functional.convex_conj` has no `_call` implementation')
        return

    f_convex_conj = functional.convex_conj
    for _ in range(100):
        y = noise_element(functional.domain)
        f_convex_conj_y = f_convex_conj(y)

        x = noise_element(functional.domain)
        lhs = x.inner(y) - functional(x)

        if not lhs <= f_convex_conj_y + EPS:
            print(repr(functional), repr(f_convex_conj), x, y, lhs,
                  f_convex_conj_y)

        assert lhs <= f_convex_conj_y + EPS


def test_proximal_convex_conj_kl_cross_entropy_solving_opt_problem():
    """Test for proximal operator of conjguate of 2nd kind KL-divergence.

    The test solves the problem

        min_x lam*KL(x | g) + 1/2||x-a||^2_2,

    where g is the nonnegative prior, and a is any vector.  Explicit solution
    to this problem is given by

        x = lam*W(g*e^(a/lam)/lam),

    where W is the Lambert W function.
    """

    # Image space
    space = odl.uniform_discr(0, 1, 10)

    # Data
    g = space.element(np.arange(10, 0, -1))
    a = space.element(np.arange(4, 14, 1))

    # Creating and assembling linear operators and proximals
    id_op = odl.IdentityOperator(space)
    lin_ops = [id_op, id_op]
    lam_kl = 2.3
    kl_ce = odl.functionals.KullbackLeiblerCrossEntropy(space, prior=g)
    g_funcs = [lam_kl * kl_ce,
               0.5 * odl.functionals.L2NormSquared(space).translated(a)]
    f = odl.functionals.ZeroFunctional(space)

    # Staring point
    x = space.zero()

    odl.solvers.douglas_rachford_pd(x, f, g_funcs, lin_ops,
                                    tau=2.1, sigma=[0.4, 0.4], niter=100)

    # Explicit solution: x = W(g * exp(a)), where W is the Lambert W function.
    x_verify = lam_kl * scipy.special.lambertw(
        (g.data / lam_kl) * np.exp(a.data / lam_kl))
    assert all_almost_equal(x, x_verify, ndigits=6)


if __name__ == '__main__':
    odl.core.util.test_file(__file__, ['-S', 'largescale'])



# Copyright 2014-2017 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

from __future__ import division
import pytest
import odl
from odl.core.util.testutils import simple_fixture, noise_element


# --- pytest fixtures --- #


hilbert_spaces = [odl.rn(3), odl.cn(3), odl.uniform_discr(0, 1, 3)]
normed_spaces = [odl.rn(3, exponent=1)] + hilbert_spaces
metric_spaces = normed_spaces
linear_spaces = metric_spaces

hilbert_space = simple_fixture('hilbert_space', hilbert_spaces)
normed_space = simple_fixture('normed_space', normed_spaces)
metric_space = simple_fixture('metric_space', metric_spaces)
linear_space = simple_fixture('linear_space', linear_spaces)


# --- LinearSpace tests --- #


def test_hash(linear_space):
    """Verify that hashing spaces works but elements doesnt."""
    hsh = hash(linear_space)

    # Check that the trivial hash algorithm is not used
    assert hsh != id(linear_space)

    x = noise_element(linear_space)
    with pytest.raises(TypeError):
        hash(x)


def test_equality(metric_space):
    """Verify that equality testing works."""
    x = noise_element(metric_space)
    y = noise_element(metric_space)

    assert x == x
    assert y == y
    assert x != y


def test_comparsion(linear_space):
    """Verify that spaces and elements in spaces cannot be compared."""
    with pytest.raises(TypeError):
        linear_space <= linear_space
    with pytest.raises(TypeError):
        linear_space < linear_space
    with pytest.raises(TypeError):
        linear_space >= linear_space
    with pytest.raises(TypeError):
        linear_space > linear_space

    x = noise_element(linear_space)
    y = noise_element(linear_space)

    
    x <= y
    
    x < y
    
    x >= y
    
    x > y


if __name__ == '__main__':
    odl.core.util.test_file(__file__)



# Copyright 2014-2020 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Helpers for discretization-related functionality.

Most functions deal with interpolation of arrays, sampling of functions and
providing a single interface for the sampler by wrapping functions or
arrays of functions appropriately.
"""

from __future__ import absolute_import, division, print_function

import inspect
import sys
from builtins import object
from functools import partial
from itertools import product   

from typing import Callable
from odl.core.set.domain import IntervalProd

import numpy as np

from odl.core.array_API_support import asarray, lookup_array_backend, ArrayBackend, get_array_and_backend
from odl.core.array_API_support.utils import is_array_supported

from odl.core.util.npy_compat import AVOID_UNNECESSARY_COPY

from odl.core.util.dtype_utils import _universal_dtype_identifier, is_floating_dtype, real_dtype, is_int_dtype
from odl.core.util import (
    dtype_repr, is_real_dtype, is_string, is_valid_input_array,
    is_valid_input_meshgrid, out_shape_from_array, out_shape_from_meshgrid,
    writable_array)

__all__ = (
    'point_collocation',
    'nearest_interpolator',
    'linear_interpolator',
    'per_axis_interpolator',
    'sampling_function',
)

SUPPORTED_INTERP = ['nearest', 'linear']


def point_collocation(func, points, out=None, **kwargs):
    """Sample a function on a grid of points.

    This function represents the simplest way of discretizing a function.
    It does little more than calling the function on a single point or a
    set of points, and returning the result.

    Parameters
    ----------
    func : callable
        Function to be sampled. It is expected to work with single points,
        meshgrids and point arrays, and to support an optional ``out``
        argument.
        Usually, ``func`` is the return value of `make_func_for_sampling`.
    points : point, meshgrid or array of points
        The point(s) where to sample.
    out : numpy.ndarray, optional
        Array to which the result should be written.
    kwargs :
        Additional arguments that are passed on to ``func``.

    Returns
    -------
    out : numpy.ndarray
        Array holding the values of ``func`` at ``points``. If ``out`` was
        given, the returned object is a reference to it.

    Examples
    --------
    Sample a 1D function:

    >>> from odl.core.discr.grid import sparse_meshgrid
    >>> domain = odl.IntervalProd(0, 5)
    >>> func = sampling_function(lambda x: x ** 2, domain, out_dtype=float)
    >>> mesh = sparse_meshgrid([1, 2, 3])
    >>> point_collocation(func, mesh)
    array([ 1.,  4.,  9.])

    By default, inputs are checked against ``domain`` to be in bounds. This
    can be switched off by passing ``bounds_check=False``:

    >>> mesh = sparse_meshgrid([-1, 0, 4])
    >>> point_collocation(func, mesh, bounds_check=False)
    array([  1.,   0.,  16.])

    In two or more dimensions, the function to be sampled can be written as
    if its arguments were the components of a point, and an implicit loop
    around the call would iterate over all points:

    >>> domain = odl.IntervalProd([0, 0], [5, 5])
    >>> xs = [1, 2]
    >>> ys = [3, 4, 5]
    >>> mesh = sparse_meshgrid(xs, ys)
    >>> func = sampling_function(lambda x: x[0] - x[1], domain, out_dtype=float)
    >>> point_collocation(func, mesh)
    array([[-2., -3., -4.],
           [-1., -2., -3.]])

    It is possible to return results that require broadcasting, and to use
    *optional* function parameters:

    >>> def f(x, c=0):
    ...     return x[0] + c
    >>> func = sampling_function(f, domain, out_dtype=float)
    >>> point_collocation(func, mesh)  # uses default c=0
    array([[ 1.,  1.,  1.],
           [ 2.,  2.,  2.]])
    >>> point_collocation(func, mesh, c=2)
    array([[ 3.,  3.,  3.],
           [ 4.,  4.,  4.]])

    The ``point_collocation`` function also supports vector- and tensor-valued
    functions. They can be given either as a single function returning an
    array-like of results, or as an array-like of member functions:

    >>> domain = odl.IntervalProd([0, 0], [5, 5])
    >>> xs = [1, 2]
    >>> ys = [3, 4]
    >>> mesh = sparse_meshgrid(xs, ys)
    >>> def vec_valued(x):
    ...     return (x[0] - 1., 0., x[0] + x[1])  # broadcasting
    >>> # For a function with several output components, we must specify the
    >>> # shape explicitly in the `out_dtype` parameter
    >>> func1 = sampling_function(
    ...     vec_valued, domain, out_dtype=float
    ... )
    >>> point_collocation(func1, mesh)
    [array([[ 0.,  0.],
           [ 1.,  1.]]), array([[ 0.,  0.],
           [ 0.,  0.]]), array([[ 4.,  5.],
           [ 5.,  6.]])]
    >>> list_of_funcs = [  # equivalent to `vec_valued`
    ...     lambda x: x[0] - 1,
    ...     0,                   # constants are allowed
    ...     lambda x: x[0] + x[1]
    ... ]

    Notes
    -----
    This function expects its input functions to be written in a
    vectorization-conforming manner to ensure fast evaluation.
    See the `ODL vectorization guide`_ for a detailed introduction.

    See Also
    --------
    make_func_for_sampling : wrap a function
    odl.core.discr.grid.RectGrid.meshgrid
    numpy.meshgrid

    References
    ----------
    .. _ODL vectorization guide:
       https://odlgroup.github.io/odl/guide/in_depth/vectorization_guide.html
    """
    if out is None:
        out = func(points, **kwargs)
    else:
        func(points, out=out, **kwargs)
    return out


def _normalize_interp(interp, ndim):
    """Turn interpolation type into a tuple with one entry per axis."""
    interp_in = interp

    if is_string(interp):
        interp = str(interp).lower()
        interp_byaxis = (interp,) * ndim
    else:
        interp_byaxis = tuple(str(itp).lower() for itp in interp)
        if len(interp_byaxis) != ndim:
            raise ValueError(
                'length of `interp` ({}) does not match number of axes ({})'
                ''.format(len(interp_byaxis), ndim)
            )

    if not all(
        interp in SUPPORTED_INTERP for interp in interp_byaxis
    ):
        raise ValueError(
            'invalid `interp` {!r}; supported are: {}'
            ''.format(interp_in, SUPPORTED_INTERP)
        )

    return interp_byaxis


def _check_interp_input(x, f):
    """Return transformed ``x``, its input type and whether it's scalar.

    On bad input, raise ``ValueError``.
    """
    errmsg_1d = (
        'bad input: expected scalar, array-like of shape (1,), (n,) or '
        '(1, n), or a meshgrid of length 1; got {!r}'
        ''.format(x)
    )

    errmsg_nd = (
        'bad input: expected scalar, array-like of shape ({0},) or '
        '({0}, n), or a meshgrid of length {0}; got {1!r}'
        ''.format(f.ndim, x)
    )

    if is_valid_input_meshgrid(x, f.ndim):
        x_is_scalar = False
        x_type = 'meshgrid'
    else:
        ### Parsing the input
        if isinstance(x, (int,float,complex, list, tuple)):
            x = np.asarray(x)
        else:
            x, _ = get_array_and_backend(x)

        if f.ndim == 1 and x.shape == ():
            x_is_scalar = True
            x = x.reshape((1, 1))
        elif f.ndim == 1 and x.ndim == 1:
            x_is_scalar = False
            x = x.reshape((1, x.size))
        elif f.ndim > 1 and x.shape == (f.ndim,):
            x_is_scalar = True
            x = x.reshape((f.ndim, 1))
        else:
            x_is_scalar = False

        if not is_valid_input_array(x, f.ndim):
            errmsg = errmsg_1d if f.ndim == 1 else errmsg_nd
            raise ValueError(errmsg)

        x_type = 'array'

    return x, x_type, x_is_scalar


def nearest_interpolator(f, coord_vecs):
    """Return the nearest neighbor interpolator for discrete values.

    Given points ``x[1] < x[2] < ... < x[N]``, and function values
    ``f[1], ..., f[N]``, nearest neighbor interpolation at ``x`` is defined
    as ::

        I(x) = f[j]

    with ``j`` such that ``|x - x[j]|`` is minimal.
    The ambiguity at the midpoints is resolved by preferring the right
    neighbor. In higher dimensions, this principle is applied per axis.
    The returned interpolator is the piecewise constant function ``x -> I(x)``.

    Parameters
    ----------
    f : numpy.ndarray
        Function values that should be interpolated.
    coord_vecs : sequence of numpy.ndarray
        Coordinate vectors of the rectangular grid on which interpolation
        should be based. They must be sorted in ascending order. Usually
        they are obtained as ``grid.coord_vectors`` from a `RectGrid`.

    Returns
    -------
    interpolator : function
        Python function that will interpolate the given values when called
        with a point or multiple points (vectorized).

    Examples
    --------
    We interpolate a 1d function. If called with a single point, the
    interpolator returns a single value, and with multiple points at once,
    an array of values is returned:

    >>> part = odl.uniform_partition(0, 2, 5)
    >>> part.coord_vectors  # grid points
    (array([ 0.2,  0.6,  1. ,  1.4,  1.8]),)
    >>> f = odl.tensor_space(5, dtype=int).element([1, 2, 3, 4, 5])
    >>> interpolator = nearest_interpolator(f, part.coord_vectors)
    >>> interpolator(0.3)  # closest to 0.2 -> value 1
    1
    >>> interpolator([0.6, 1.3, 1.9])  # closest to [0.6, 1.4, 1.8]
    array([2, 4, 5], dtype=int32)

    In 2 dimensions, we can either use a (transposed) list of points or
    a meshgrid:

    >>> part = odl.uniform_partition([0, 0], [1, 5], shape=(2, 4))
    >>> part.coord_vectors  # grid points
    (array([ 0.25,  0.75]), array([ 0.625,  1.875,  3.125,  4.375]))
    >>> f = np.array([[1, 2, 3, 4],
    ...               [5, 6, 7, 8]],
    ...              dtype=float)
    >>> interpolator = nearest_interpolator(f, part.coord_vectors)
    >>> interpolator([1, 1])  # single point
    5.0
    >>> x = np.array([[0.5, 2.0],
    ...               [0.0, 4.5],
    ...               [0.0, 3.0]]).T  # 3 points at once
    >>> interpolator(x)
    array([ 6.,  4.,  3.])
    >>> from odl.core.discr.grid import sparse_meshgrid
    >>> mesh = sparse_meshgrid([0.0, 0.4, 1.0], [1.5, 3.5])
    >>> interpolator(mesh)  # 3x2 grid of points
    array([[ 2.,  3.],
           [ 2.,  3.],
           [ 6.,  7.]])


    See Also
    --------
    linear_interpolator : (bi-/tri-/...)linear interpolation
    per_axis_interpolator : potentially different interpolation in each axis

    Notes
    -----
    - **Important:** if called on a point array, the points are
      assumed to be sorted in ascending order in each dimension
      for efficiency reasons.
    - Nearest neighbor interpolation is the only scheme which works
      with data of non-numeric data type since it does not involve any
      arithmetic operations on the values, in contrast to other
      interpolation methods.
    """
    # f = np.asarray(f)
    f, backend = get_array_and_backend(f)
    # TODO(kohr-h): pass reasonable options on to the interpolator
    def nearest_interp(x, out=None):
        """Interpolating function with vectorization."""
        x, x_type, x_is_scalar = _check_interp_input(x, f)
        interpolator = _NearestInterpolator(coord_vecs, f, input_type=x_type)

        res = interpolator(x, out=out)
        if x_is_scalar:
            res = res.item()
        return res

    return nearest_interp


def linear_interpolator(f, coord_vecs):
    """Return the linear interpolator for discrete function values.

    Parameters
    ----------
    f : numpy.ndarray
        Function values that should be interpolated.
    coord_vecs : sequence of numpy.ndarray
        Coordinate vectors of the rectangular grid on which interpolation
        should be based. They must be sorted in ascending order. Usually
        they are obtained as ``grid.coord_vectors`` from a `RectGrid`.

    Returns
    -------
    interpolator : function
        Python function that will interpolate the given values when called
        with a point or multiple points (vectorized).

    Examples
    --------
    We interpolate a 1d function. If called with a single point, the
    interpolator returns a single value, and with multiple points at once,
    an array of values is returned:

    >>> part = odl.uniform_partition(0, 2, 5)
    >>> part.coord_vectors  # grid points
    (array([ 0.2,  0.6,  1. ,  1.4,  1.8]),)
    >>> f = np.array([1.0, 2.0, 3.0, 4.0, 5.0])
    >>> interpolator = linear_interpolator(f, part.coord_vectors)
    >>> interpolator(0.3)  # 0.75 * 1 + 0.25 * 2 = 1.25
    1.25
    >>> # At 1.9, the value is interpolated between the last value 5.0 and
    >>> # 0.0. The extra interpolation node is placed at the same distance
    >>> # as the second-to-last, i.e., at 2.2. Hence, the interpolated value
    >>> # is 0.75 * 5.0 + 0.25 * 0.0 = 3.75.
    >>> interpolator([0.6, 1.3, 1.9])
    array([ 2.  ,  3.75,  3.75])

    In 2 dimensions, we can either use a (transposed) list of points or
    a meshgrid:

    >>> part = odl.uniform_partition([0, 0], [1, 5], shape=(2, 4))
    >>> part.coord_vectors  # grid points
    (array([ 0.25,  0.75]), array([ 0.625,  1.875,  3.125,  4.375]))
    >>> f = np.array([[1, 2, 3, 4],
    ...               [5, 6, 7, 8]],
    ...              dtype=float)
    >>> interpolator = linear_interpolator(f, part.coord_vectors)
    >>> interpolator([1, 1])  # single point
    2.65
    >>> x = np.array([[0.5, 2.0],
    ...               [0.0, 4.5],
    ...               [0.0, 3.0]]).T  # 3 points at once
    >>> interpolator(x)
    array([ 4.1 ,  1.8 ,  1.45])
    >>> from odl.core.discr.grid import sparse_meshgrid
    >>> mesh = sparse_meshgrid([0.0, 0.5, 1.0], [1.5, 3.5])
    >>> interpolator(mesh)  # 3x2 grid of points
    array([[ 0.85,  1.65],
           [ 3.7 ,  5.3 ],
           [ 2.85,  3.65]])
    """
    f = asarray(f)

    # TODO(kohr-h): pass reasonable options on to the interpolator
    def linear_interp(x, out=None):
        """Interpolating function with vectorization."""
        x, x_type, x_is_scalar = _check_interp_input(x, f)
        interpolator = _LinearInterpolator(
            coord_vecs,
            f,
            input_type=x_type
        )

        res = interpolator(x, out=out)
        if x_is_scalar:
            res = res.item()
        return res

    return linear_interp


def per_axis_interpolator(f, coord_vecs, interp):
    """Return a per axis defined interpolator for discrete values.

    With this function, the interpolation scheme can be chosen for each axis
    separately.

    Parameters
    ----------
    f : numpy.ndarray
        Function values that should be interpolated.
    coord_vecs : sequence of numpy.ndarray
        Coordinate vectors of the rectangular grid on which interpolation
        should be based. They must be sorted in ascending order. Usually
        they are obtained as ``grid.coord_vectors`` from a `RectGrid`.
    interp : str or sequence of str
        Indicates which interpolation scheme to use for which axis.
        A single string is interpreted as a global scheme for all
        axes.

    Examples
    --------
    Choose linear interpolation in the first axis and nearest neighbor in
    the second:

    >>> part = odl.uniform_partition([0, 0], [1, 5], shape=(2, 4))
    >>> part.coord_vectors
    (array([ 0.25,  0.75]), array([ 0.625,  1.875,  3.125,  4.375]))
    >>> f = np.array([[1, 2, 3, 4],
    ...               [5, 6, 7, 8]],
    ...              dtype=float)
    >>> interpolator = per_axis_interpolator(
    ...     f, part.coord_vectors, ['linear', 'nearest']
    ... )
    >>> interpolator([1, 1])  # single point
    2.5
    >>> x = np.array([[0.5, 2.0],
    ...               [0.0, 4.5],
    ...               [0.0, 3.0]]).T  # 3 points at once
    >>> interpolator(x)
    array([ 4. ,  2. ,  1.5])
    >>> from odl.core.discr.grid import sparse_meshgrid
    >>> mesh = sparse_meshgrid([0.0, 0.5, 1.0], [1.5, 3.5])
    >>> interpolator(mesh)  # 3x2 grid of points
    array([[ 1. ,  1.5],
           [ 4. ,  5. ],
           [ 3. ,  3.5]])
    """
    f = asarray(f)

    interp = _normalize_interp(interp, f.ndim)

    def per_axis_interp(x, out=None):
        """Interpolating function with vectorization."""
        x, x_type, x_is_scalar = _check_interp_input(x, f)
        interpolator = _PerAxisInterpolator(
            coord_vecs, f, interp=interp, input_type=x_type
        )

        res = interpolator(x, out=out)
        if x_is_scalar:
            res = res.item()
        return res

    return per_axis_interp


class _Interpolator(object):
    r"""Abstract interpolator class.

    The code is adapted from SciPy's `RegularGridInterpolator
    <http://docs.scipy.org/doc/scipy/reference/generated/scipy.interpolate.RegularGridInterpolator.html>`_
    class.

    The init method does not convert to floating point to
    support arbitrary data type for nearest neighbor interpolation.

    Subclasses need to override ``_evaluate`` for concrete
    implementations.
    """

    def __init__(self, coord_vecs, values, input_type):
        """Initialize a new instance.

        coord_vecs : sequence of `numpy.ndarray`'s
            Coordinate vectors defining the interpolation grid.
        values : `array-like`
            Grid values to use for interpolation.
        input_type : {'array', 'meshgrid'}
            Type of expected input values in ``__call__``.
        """
        values, backend = get_array_and_backend(values)
        typ_ = str(input_type).lower()
        if typ_ not in ('array', 'meshgrid'):
            raise ValueError('`input_type` ({}) not understood'
                             ''.format(input_type))

        if len(coord_vecs) != values.ndim:
            raise ValueError(
                'there are {} point arrays, but `values` has {} dimensions'
                ''.format(len(coord_vecs), values.ndim)
            )
        for i, p in enumerate(coord_vecs):
            if not np.asarray(p).ndim == 1:
                raise ValueError('the points in dimension {} must be '
                                 '1-dimensional'.format(i))
            if values.shape[i] != len(p):
                raise ValueError('there are {} points and {} values in '
                                 'dimension {}'.format(len(p),
                                                       values.shape[i], i))

        self.coord_vecs = tuple(np.asarray(p) for p in coord_vecs)
        self.values = values
        self.input_type = input_type

        self.backend = backend
        self.namespace = backend.array_namespace
        self.device = values.device

    def __call__(self, x, out=None):
        """Do the interpolation.

        Parameters
        ----------
        x : `meshgrid` (i.e., tuple of arrays) if `input_type` is meshgrid,
            else `numpy.ndarray`.
            Evaluation points of the interpolator.
        out : `numpy.ndarray`, optional
            Array to which the results are written. Needs to have
            correct shape according to input ``x``.

        Returns
        -------
        out : `numpy.ndarray`
            Interpolated values. If ``out`` was given, the returned
            object is a reference to it.
        """
        if self.input_type == 'meshgrid':
            # Given a meshgrid, the evaluation will be on a ragged array.
            x = [get_array_and_backend(x_)[0] for x_ in x]
        else:
            x = get_array_and_backend(x)[0]

        ndim = len(self.coord_vecs)
        scalar_out = False

        if self.input_type == 'array':
            if ndim == 1:
                scalar_out = x.ndim == 0
            else:
                scalar_out = x.shape == (ndim,)
            # Make a (1, n) array from one with shape (n,)
            x = x.reshape([ndim, -1])
            out_shape = out_shape_from_array(x)
        else:
            if len(x) != ndim:
                raise ValueError('number of vectors in x is {} instead of '
                                 'the grid dimension {}'
                                 ''.format(len(x), ndim))
            out_shape = out_shape_from_meshgrid(x)

        if out is not None:
            if not isinstance(out, self.backend.array_type):
                raise TypeError(f'The provided out argument is not an expected {type(self.backend.array_type)} but a {type(out)}')
            if out.shape != out_shape:
                raise ValueError('output shape {} not equal to expected '
                                 'shape {}'.format(out.shape, out_shape))
            if out.dtype != self.values.dtype:
                raise ValueError('output dtype {} not equal to expected '
                                 'dtype {}'
                                 ''.format(out.dtype, self.values.dtype))

        indices, norm_distances = self._find_indices(x)
        values = self._evaluate(indices, norm_distances, out)
        if scalar_out:
            return values.item()
        else:
            return values

    def _find_indices(self, x):
        """Find indices and distances of the given nodes.

        Can be overridden by subclasses to improve efficiency.
        """
        # find relevant edges between which xi are situated
        index_vecs = []
        # compute distance to lower edge in unity units
        norm_distances = []

        # iterate through dimensions
        for xi, cvec in zip(x, self.coord_vecs):
            # try:
            if is_floating_dtype(self.values.dtype):
                dtype = real_dtype(self.values.dtype, backend=self.backend)
            elif is_int_dtype(self.values.dtype):
                dtype = real_dtype(float, backend=self.backend)
            else:
                raise ValueError(f'Values can only be integers or float, not {type(self.values)}')
            xi = self.backend.array_constructor(xi, dtype = dtype, device=self.device)
            cvec = self.backend.array_constructor(cvec, dtype = dtype, device=self.device)
            idcs = self.namespace.searchsorted(cvec, xi) - 1
            idcs[idcs < 0] = 0
            idcs[idcs > len(cvec) - 2] = len(cvec) - 2
            index_vecs.append(idcs)

            try:
             norm_distances.append((xi - cvec[idcs]) /
                                  (cvec[idcs + 1] - cvec[idcs]))
            except Exception as e:
                print(f"{type(xi)=}, {type(cvec)=}")
                raise e

        return index_vecs, norm_distances

    def _evaluate(self, indices, norm_distances, out=None):
        """Evaluation method, needs to be overridden."""
        raise NotImplementedError('abstract method')


class _NearestInterpolator(_Interpolator):

    """Nearest neighbor interpolator.

    The code is adapted from SciPy's `RegularGridInterpolator
    <http://docs.scipy.org/doc/scipy/reference/generated/\
scipy.interpolate.RegularGridInterpolator.html>`_ class.

    This implementation is faster than the more generic one in the
    `_PerAxisPointwiseInterpolator`.
    """

    def __init__(self, coord_vecs, values, input_type):
        """Initialize a new instance.

        coord_vecs : sequence of `numpy.ndarray`'s
            Coordinate vectors defining the interpolation grid
        values : `array-like`
            Grid values to use for interpolation
        input_type : {'array', 'meshgrid'}
            Type of expected input values in ``__call__``
        """
        super(_NearestInterpolator, self).__init__(
            coord_vecs, values, input_type
        )

    def _evaluate(self, indices, norm_distances, out=None):
        """Evaluate nearest interpolation."""
        idx_res = []
        
        for i, yi in zip(indices, norm_distances):            
            idx_res.append(self.namespace.where(yi < .5, i, i + 1))
        idx_res = tuple(idx_res)
        if out is not None:
            out[:] = self.values[idx_res]
            return out
        else:
            return self.values[idx_res]


def _compute_nearest_weights_edge(idcs, ndist, backend):
    """Helper for nearest interpolation mimicing the linear case."""
    # Get out-of-bounds indices from the norm_distances. Negative
    # means "too low", larger than or equal to 1 means "too high"
    lo = (ndist < 0)
    hi = (ndist > 1)

    # For "too low" nodes, the lower neighbor gets weight zero;
    # "too high" gets 1.
    w_lo = backend.array_namespace.where(ndist < 0.5, 1.0, 0.0)
    w_lo[lo] = 0
    w_lo[hi] = 1

    # For "too high" nodes, the upper neighbor gets weight zero;
    # "too low" gets 1.
    w_hi = backend.array_namespace.where(ndist < 0.5, 0.0, 1.0)
    w_hi[lo] = 1
    w_hi[hi] = 0

    # For upper/lower out-of-bounds nodes, we need to set the
    # lower/upper neighbors to the last/first grid point
    edge = [idcs, idcs + 1]
    edge[0][hi] = -1
    edge[1][lo] = 0

    return w_lo, w_hi, edge


def _compute_linear_weights_edge(idcs, ndist, backend):
    """Helper for linear interpolation."""
    assert(isinstance(ndist, backend.array_type))

    # Get out-of-bounds indices from the norm_distances. Negative
    # means "too low", larger than or equal to 1 means "too high"
    if backend.impl == 'numpy': 
        lo = backend.array_namespace.where(ndist < 0, ndist, 0).nonzero()
        hi = backend.array_namespace.where(ndist > 1, ndist, 0).nonzero()
    elif backend.impl == 'pytorch': 
        lo = backend.array_namespace.where(ndist < 0, ndist, 0).nonzero(as_tuple=True)
        hi = backend.array_namespace.where(ndist > 1, ndist, 0).nonzero(as_tuple=True)
    else:
        raise NotImplementedError
    # For "too low" nodes, the lower neighbor gets weight zero;
    # "too high" gets 2 - yi (since yi >= 1)
    w_lo = (1 - ndist)
    w_lo[lo] = 0
    w_lo[hi] += 1

    # For "too high" nodes, the upper neighbor gets weight zero;
    # "too low" gets 1 + yi (since yi < 0)
    w_hi = backend.array_constructor(ndist, copy=True)
    w_hi[lo] += 1
    w_hi[hi] = 0

    # For upper/lower out-of-bounds nodes, we need to set the
    # lower/upper neighbors to the last/first grid point
    edge = [idcs, idcs + 1]
    edge[0][hi] = -1
    edge[1][lo] = 0

    return w_lo, w_hi, edge


def _create_weight_edge_lists(indices, norm_distances, interp, backend):
    # Pre-calculate indices and weights (per axis)
    low_weights = []
    high_weights = []
    edge_indices = []
    for i, (idcs, yi, s) in enumerate(zip(indices, norm_distances, interp)):
        if s == 'nearest':
            w_lo, w_hi, edge = _compute_nearest_weights_edge(idcs, yi, backend=backend)
        elif s == 'linear':
            w_lo, w_hi, edge = _compute_linear_weights_edge(idcs, yi, backend=backend)
        else:
            raise ValueError('invalid `interp` {}'.format(interp))

        low_weights.append(w_lo)
        high_weights.append(w_hi)
        edge_indices.append(edge)

    return low_weights, high_weights, edge_indices


class _PerAxisInterpolator(_Interpolator):
    """Interpolator where the scheme is set per axis.

    This allows to use e.g. nearest neighbor interpolation in the
    first dimension and linear in dimensions 2 and 3.
    """

    def __init__(self, coord_vecs, values, input_type, interp):
        """Initialize a new instance.

        coord_vecs : sequence of `numpy.ndarray`'s
            Coordinate vectors defining the interpolation grid
        values : `array-like`
            Grid values to use for interpolation
        input_type : {'array', 'meshgrid'}
            Type of expected input values in ``__call__``
        interp : sequence of str
            Indicates which interpolation scheme to use for which axis
        """
        super(_PerAxisInterpolator, self).__init__(
            coord_vecs, values, input_type)
        self.interp = interp

    def _evaluate(self, indices, norm_distances, out=None):
        """Evaluate per-axis interpolation.

        Modified for in-place evaluation and treatment of out-of-bounds
        points by implicitly assuming 0 at the next node.
        """
        # slice for broadcasting over trailing dimensions in self.values
        vslice = (slice(None),) + (None,) * (self.values.ndim - len(indices))

        if out is None:
            out_shape = out_shape_from_meshgrid(norm_distances)
            out_dtype = self.values.dtype
            out = self.namespace.zeros(
                out_shape, dtype=out_dtype, device=self.device
                )
        else:
            out[:] = 0.0

        # Weights and indices (per axis)
        low_weights, high_weights, edge_indices = _create_weight_edge_lists(
            indices, norm_distances, self.interp, backend=self.backend)

        # Iterate over all possible combinations of [i, i+1] for each
        # axis, resulting in a loop of length 2**ndim
        for lo_hi, edge in zip(product(*([['l', 'h']] * len(indices))),
                               product(*edge_indices)):
            weight = self.backend.array_constructor(
                [1.0], dtype=self.values.dtype, device=self.device
                )
            # TODO(kohr-h): determine best summation order from array strides
            for lh, w_lo, w_hi in zip(lo_hi, low_weights, high_weights):

                # We don't multiply in-place to exploit the cheap operations
                # in the beginning: sizes grow gradually as following:
                # (n, 1, 1, ...) -> (n, m, 1, ...) -> ...
                # Hence, it is faster to build up the weight array instead
                # of doing full-size operations from the beginning.
                # Emilien : This array-API compatibility is horribly slow ( sending the individual floats to the gpu while iterating is a hack around the inhomogeneous dimensions returned by _create_weight_edge_lists)
                if lh == 'l':
                    weight = weight * self.backend.array_constructor(
            w_lo, device=self.device)
                else:
                    weight = weight * self.backend.array_constructor(
            w_hi, device=self.device)
            out += self.backend.array_constructor(self.values[edge], device=self.device) * weight[vslice]
        # return np.array(out, copy=AVOID_UNNECESSARY_COPY, ndmin=1)
        return self.backend.array_constructor(
            out, copy=AVOID_UNNECESSARY_COPY, device=self.device
            )


class _LinearInterpolator(_PerAxisInterpolator):
    """Linear (i.e. bi-/tri-/multi-linear) interpolator.

    Convenience class.
    """

    def __init__(self, coord_vecs, values, input_type):
        """Initialize a new instance.

        coord_vecs : sequence of `numpy.ndarray`'s
            Coordinate vectors defining the interpolation grid
        values : `array-like`
            Grid values to use for interpolation
        input_type : {'array', 'meshgrid'}
            Type of expected input values in ``__call__``
        """
        super(_LinearInterpolator, self).__init__(
            coord_vecs,
            values,
            input_type,
            interp=['linear'] * len(coord_vecs),
        )


def _check_func_out_arg(func):
    """Check if ``func`` has an (optional) ``out`` argument.

    Also verify that the signature of ``func`` has no ``*args`` since
    they make argument propagation a huge hassle.

    Note: this function only works for objects that can be inspected
    with the ``inspect`` module, i.e., Python functions and callables,
    but not, e.g., NumPy UFuncs.

    Parameters
    ----------
    func : callable
        Object that should be inspected.

    Returns
    -------
    has_out : bool
        ``True`` if the signature has an ``out`` argument, ``False``
        otherwise.
    out_is_optional : bool
        ``True`` if ``out`` is present and optional in the signature,
        ``False`` otherwise.

    Raises
    ------
    TypeError
        If ``func``'s signature has ``*args``.
    """
    if sys.version_info.major > 2:
        spec = inspect.getfullargspec(func)
        kw_only = spec.kwonlyargs
    else:
        spec = inspect.getargspec(func)
        kw_only = ()

    if spec.varargs is not None:
        raise TypeError('*args not allowed in function signature')

    pos_args = spec.args
    pos_defaults = () if spec.defaults is None else spec.defaults

    if 'out' in pos_args:
        has_out = True
        out_optional = (
            pos_args.index('out') >= len(pos_args) - len(pos_defaults)
        )
    elif 'out' in kw_only:
        has_out = out_optional = True
    else:
        has_out = out_optional = False

    return has_out, out_optional


def _func_out_type(func):
    """Determine the output argument type (if any) of a function-like object.

    This function is intended to work with all types of callables
    that are used as input to `sampling_function`.
    """
    # Numpy `UFuncs` and similar objects (e.g. Numba `DUFuncs`)
    if hasattr(func, 'nin') and hasattr(func, 'nout'):
        if func.nin != 1:
            raise ValueError(
                'ufunc {} takes {} input arguments, expected 1'
                ''.format(func.__name__, func.nin)
            )
        if func.nout > 1:
            raise ValueError(
                'ufunc {} returns {} outputs, expected 0 or 1'
                ''.format(func.__name__, func.nout)
            )
        has_out = out_optional = (func.nout == 1)
    elif inspect.isfunction(func):
        has_out, out_optional = _check_func_out_arg(func)
    elif callable(func):
        has_out, out_optional = _check_func_out_arg(func.__call__)
    else:
        raise TypeError('object {!r} not callable'.format(func))

    return has_out, out_optional


def _broadcast_nested_list(arr_lists, element_shape, ndim, backend: ArrayBackend):
    """ A generalisation of `np.broadcast_to`, applied to an arbitrarily
    deep list (or tuple) eventually containing arrays or scalars. """
    if isinstance(arr_lists, backend.array_type) or np.isscalar(arr_lists):
        if ndim == 1:
            # As usual, 1d is tedious to deal with. This
            # code deals with extra dimensions in result
            # components that stem from using x instead of
            # x[0] in a function.
            # Without this, broadcasting fails.
            shp = getattr(arr_lists, 'shape', ())
            if shp and shp[0] == 1:
                arr_lists = arr_lists.reshape(arr_lists.shape[1:])
        return backend.array_namespace.broadcast_to(arr_lists, element_shape)
    else:
        return [_broadcast_nested_list(row, element_shape, ndim, backend=backend)
                 for row in arr_lists]

def _send_nested_list_to_backend(
        arr_lists, backend : ArrayBackend, device, dtype
        ):
    if backend.impl == 'numpy':
        return arr_lists

    if isinstance(arr_lists, np.ndarray) or np.isscalar(arr_lists):
        return backend.array_constructor(arr_lists, device=device, dtype=dtype)

    elif isinstance(arr_lists, (tuple,list)):
        return [_send_nested_list_to_backend(arr, backend, device, dtype) for arr in arr_lists]

    else:
        raise TypeError(f'Type of input {type(arr_lists)} not supported.')

def sampling_function(
    func : Callable | list | tuple,
    domain : IntervalProd,
    out_dtype : str = None,
    impl: str ='numpy',
    device: str ='cpu'
    ):
    def _infer_dtype(out_dtype : str | None):
        if out_dtype is None:
            out_dtype = 'float64'
        else:
            assert is_floating_dtype(out_dtype)
        return out_dtype

    def _sanitise_callable(func: Callable) -> Callable:
        # Get default implementations if necessary
        has_out, out_optional = _func_out_type(func)

        if has_out:
            raise NotImplementedError('Currently, not implemented for out-of-place functions')

        return func

    def _sanitise_input_function(func: Callable):
        '''
        This function aims at unpacking the input function `func`.
        The former API expects a callable or array-like (of callables)
        The new API checks 
        '''
        if isinstance(func, Callable):
            return _sanitise_callable(func)
        elif isinstance(func, (list, tuple)):
            raise NotImplementedError('The sampling function cannot be instantiated with a list-like of callables.')
        else:
            raise NotImplementedError('The function to sample must be either a Callable or an array-like (list, tuple) of callables.')

    ### We begin by sanitising the inputs:
    # 1) the dtype
    out_dtype = _infer_dtype(out_dtype)
    # 2) the func_or_arr
    func = _sanitise_input_function(func)

    ### We then create the function
    return _make_single_use_func(func, domain, out_dtype, impl, device)



def old_sampling_function(func_or_arr, domain, out_dtype=None, impl: str ='numpy', device: str ='cpu'):
    """Return a function that can be used for sampling.

    For examples on this function's usage, see `point_collocation`.

    Parameters
    ----------
    func_or_arr : callable or array-like
        Either a single callable object (possibly with multiple output
        components), or an array or callables and constants.
        A callable (or each callable) must take a single input and may
        accept one output parameter called ``out``, and should return
        its result.
    domain : IntervalProd
        Set in which inputs to the function are assumed to lie. It is used
        to determine the type of input (point/meshgrid/array) based on
        ``domain.ndim``, and (unless switched off) to check whether all
        inputs are in bounds.
    out_dtype : optional
        Data type of a *single* output of ``func_or_arr``, i.e., when
        called with a single point as input. In particular:

        - If ``func_or_arr`` is a scalar-valued function, ``out_dtype`` is
          expected to be a basic dtype with empty shape.
        - If ``func_or_arr`` is a vector- or tensor-valued function,
          ``out_dtype`` should be a shaped data type, e.g., ``(float, (3,))``
          for a vector-valued function with 3 components.
        - If ``func_or_arr`` is an array-like, ``out_dtype`` should be a
          shaped dtype whose shape matches that of ``func_or_arr``. It can
          also be ``None``, in which case the shape is inferred, and the
          scalar data type is set to ``float``.

    Returns
    -------
    func : function
        Wrapper function that has no optional ``out`` argument.
    """
    if out_dtype is None:
        val_shape = None
        scalar_out_dtype = np.dtype('float64')
    else:
        out_dtype = np.dtype(out_dtype)
        val_shape = out_dtype.shape
        scalar_out_dtype = out_dtype.base

    # Provide default implementations of missing function signature types

    def _default_oop(func_ip, x, **kwargs):
        """Default out-of-place variant of an in-place-only function."""
        if is_valid_input_array(x, domain.ndim):
            scalar_out_shape = out_shape_from_array(x)
        elif is_valid_input_meshgrid(x, domain.ndim):
            scalar_out_shape = out_shape_from_meshgrid(x)
        else:
            raise TypeError('invalid input `x`')

        out_shape = val_shape + scalar_out_shape
        out = np.empty(out_shape, dtype=scalar_out_dtype)
        func_ip(x, out=out, **kwargs)
        return out

    def _default_ip(func_oop, x, out, **kwargs):
        """Default in-place variant of an out-of-place-only function."""
        result = func_oop(x, **kwargs)
        try:
            result = np.array(result, copy=AVOID_UNNECESSARY_COPY)
        except ValueError:
            # Different shapes encountered, need to broadcast
            if is_valid_input_array(x, domain.ndim):
                scalar_out_shape = out_shape_from_array(x)
            elif is_valid_input_meshgrid(x, domain.ndim):
                scalar_out_shape = out_shape_from_meshgrid(x)
            else:
                raise TypeError('invalid input `x`')

            bcast_results = _broadcast_nested_list(
                                   result, scalar_out_shape, domain.ndim)

            # New array that is flat in the `out_shape` axes, reshape it
            # to the final `out_shape + scalar_shape`, using the same
            # order ('C') as the initial `result.ravel()`.
            result = np.array(bcast_results, dtype=scalar_out_dtype)
            result = result.reshape(val_shape + scalar_out_shape)

        # The following code is required to remove extra axes, e.g., when
        # the result has shape (2, 1, 3) but should have shape (2, 3).
        # For those cases, broadcasting doesn't apply.
        try:
            reshaped = result.reshape(out.shape)
        except ValueError:
            # This is the case when `result` must be broadcast
            out[:] = result
        else:
            out[:] = reshaped

        return out

    # Now prepare the in-place and out-of-place functions for the final
    # wrapping.

    if callable(func_or_arr):
        # Assume scalar float out dtype for single function
        if out_dtype is None:
            out_dtype = np.dtype('float64')

        # Got a (single) function, possibly need to vectorize
        func = func_or_arr

        # Get default implementations if necessary
        has_out, out_optional = _func_out_type(func)
        if not has_out:
            # Out-of-place-only
            func_oop = func
        elif out_optional:
            # Dual-use
            func_oop = func
        else:
            # In-place-only
            func_oop = partial(_default_oop, func)

    else:
        # This is for the case that an array-like of callables is provided.
        # We need to convert this into a single function that returns an
        # array.

        arr = np.array(func_or_arr, dtype=object)

        if val_shape is None:
            # Infer value shape if `out_dtype is None`
            val_shape = arr.shape
        elif arr.shape != val_shape:
            # Otherwise, check that the value shape matches the dtype shape
            raise ValueError(
                'invalid `func_or_arr` {!r}: expected `None`, a callable or '
                'an array-like of callables whose shape matches '
                '`out_dtype.shape` {}'.format(func_or_arr, val_shape)
            )

        out_dtype = np.dtype((scalar_out_dtype, val_shape))

        arr = arr.ravel().tolist()

        def array_wrapper_func(x, out=None, **kwargs):
            """Function wrapping an array of callables and constants.

            This wrapper does the following for out-of-place
            evaluation (when ``out=None``):

            1. Collect the results of all function evaluations into
               a list, handling all kinds of sequence entries
               (normal function, ufunc, constant, etc.).
            2. Broadcast all results to the desired shape that is
               determined by the space's ``out_shape`` and the
               shape(s) of the input.
            3. Form a big array containing the final result.

            The in-place version is simpler because broadcasting
            happens automatically when assigning to the components
            of ``out``. Hence, we only have

            1. Assign the result of the evaluation of the i-th
               function to ``out_flat[i]``, possibly using the
               ``out`` parameter of the function.
            """
            if is_valid_input_meshgrid(x, domain.ndim):
                scalar_out_shape = out_shape_from_meshgrid(x)
            elif is_valid_input_array(x, domain.ndim):
                scalar_out_shape = out_shape_from_array(x)
            else:
                raise RuntimeError('bad input')

            if out is None:
                # Out-of-place evaluation

                # Collect results of member functions into a list.
                # Put simply, all that happens here is
                # `results.append(f(x))`, just for a bunch of cases
                # and with or without `out`.
                results = []
                for f in arr:
                    if np.isscalar(f):
                        # Constant function
                        results.append(f)
                    elif not callable(f):
                        raise TypeError(
                            'element {!r} of `func_or_arr` not callable'
                            ''.format(f)
                        )
                    elif hasattr(f, 'nin') and hasattr(f, 'nout'):
                        # ufunc-like object
                        results.append(f(x, **kwargs))
                    else:
                        has_out, _ = _func_out_type(f)
                        if has_out:
                            out = np.empty(
                                scalar_out_shape, dtype=scalar_out_dtype
                            )
                            f(x, out=out, **kwargs)
                            results.append(out)
                        else:
                            results.append(f(x, **kwargs))

                # Broadcast to required shape and convert to array.
                # This will raise an error if the shape of some member
                # array is wrong, since in that case the resulting
                # dtype would be `object`.
                bcast_results = []
                for res in results:
                    try:
                        reshaped = np.reshape(res, scalar_out_shape)
                    except ValueError:
                        bcast_results.append(
                            np.broadcast_to(res, scalar_out_shape))
                    else:
                        bcast_results.append(reshaped)

                out_arr = np.array(
                    bcast_results, dtype=scalar_out_dtype
                )

                return out_arr.reshape(val_shape + scalar_out_shape)

            else:
                # In-place evaluation

                # This is a precaution in case out is not contiguous
                with writable_array(out) as out_arr:
                    # Flatten tensor axes to work on one tensor
                    # component (= scalar function) at a time
                    out_comps = out_arr.reshape((-1,) + scalar_out_shape)
                    for f, out_comp in zip(arr, out_comps):
                        if np.isscalar(f):
                            out_comp[:] = f
                        else:
                            has_out, _ = _func_out_type(f)
                            if has_out:
                                f(x, out=out_comp, **kwargs)
                            else:
                                out_comp[:] = f(x, **kwargs)

        func_oop = array_wrapper_func

    return _make_single_use_func(func_oop, domain, out_dtype, impl=impl, device=device)


def _make_single_use_func(func_oop, domain, out_dtype, impl: str ='numpy', device: str ='cpu'):
    """Return a unifying wrapper function with optional ``out`` argument."""

    # Default to `ndim=1` for unusual domains that do not define a dimension
    # (like `Strings(3)`)
    ndim = getattr(domain, 'ndim', 1)
    if out_dtype is None:
        # Don't let `np.dtype` convert `None` to `float64`
        raise TypeError('`out_dtype` cannot be `None`')

    out_dtype = np.dtype(out_dtype)
    val_shape = out_dtype.shape
    scalar_out_dtype = out_dtype.base

    tensor_valued = val_shape != ()

    def dual_use_func(x, **kwargs):
        """Wrapper function with optional ``out`` argument.

        This function closes over two other functions, one for in-place,
        the other for out-of-place evaluation. Its purpose is to unify their
        interfaces to a single one with optional ``out`` argument, and to
        automate all details of input/output checking, broadcasting and
        type casting.

        The closure also contains ``domain``, an `IntervalProd` where points
        should lie, and the expected ``out_dtype``.

        For usage examples, see `point_collocation`.

        Parameters
        ----------
        x : point, `meshgrid` or `numpy.ndarray`
            Input argument for the function evaluation. Conditions
            on ``x`` depend on its type:

            - point: must be castable to an element of the enclosed ``domain``.
            - meshgrid: length must be ``domain.ndim``, and the arrays must
              be broadcastable against each other.
            - array: shape must be ``(ndim, N)``, where ``ndim`` equals
              ``domain.ndim``.

        out : `numpy.ndarray`, optional
            Output argument holding the result of the function evaluation.
            Its shape must be ``out_dtype.shape + np.broadcast(*x).shape``.

        Other Parameters
        ----------------
        bounds_check : bool, optional
            If ``True``, check if all input points lie in ``domain``. This
            requires ``domain`` to implement `Set.contains_all`.
            Default: ``True``

        Returns
        -------
        out : `numpy.ndarray`
            Result of the function evaluation. If ``out`` was provided,
            the returned object is a reference to it.

        Raises
        ------
        TypeError
            If ``x`` is not a valid vectorized evaluation argument.

            If ``out`` is neither ``None`` nor a `numpy.ndarray` of
            adequate shape and data type.

        ValueError
            If ``bounds_check == True`` and some evaluation points fall
            outside the valid domain.
        """
        bounds_check = kwargs.pop('bounds_check', True)
        if bounds_check and not hasattr(domain, 'contains_all'):
            raise AttributeError(
                'bounds check not possible for domain {!r}, missing '
                '`contains_all()` method'
                ''.format(domain)
            )

        # Check for input type and determine output shape
        if is_valid_input_meshgrid(x, ndim):
            scalar_in = False
            scalar_out_shape = out_shape_from_meshgrid(x)

            # Avoid operations on tuples like x * 2 by casting to array
            if ndim == 1:
                x = x[0][None, ...]
        elif is_valid_input_array(x, ndim):
            x = np.asarray(x)
            scalar_in = False
            scalar_out_shape = out_shape_from_array(x)

        elif x in domain:
            x = np.atleast_2d(x).T  # make a (d, 1) array
            scalar_in = True
            scalar_out_shape = (1,)

        else:
            # Unknown input
            txt_1d = ' or (n,)' if ndim == 1 else ''
            raise TypeError(
                'argument {!r} not a valid function input. '
                'Expected an element of the domain {domain!r}, an array-like '
                'with shape ({domain.ndim}, n){} or a length-{domain.ndim} '
                'meshgrid tuple.'
                ''.format(x, txt_1d, domain=domain)
            )

        # Check bounds if specified
        if bounds_check and not domain.contains_all(x):
            raise ValueError('input contains points outside the domain {!r}'
                             ''.format(domain))

        backend = lookup_array_backend(impl)
        array_ns = backend.array_namespace
        backend_scalar_out_dtype = backend.available_dtypes[_universal_dtype_identifier(scalar_out_dtype)]

        x = _send_nested_list_to_backend(
            x, backend, device, backend_scalar_out_dtype)

        if scalar_in:
            out_shape = val_shape
        else:
            out_shape = val_shape + scalar_out_shape

        if ndim == 1:
            try:
                out = func_oop(x, **kwargs)
            except (TypeError, IndexError):
                # TypeError is raised if a meshgrid was used but the
                # function expected an array (1d only). In this case we try
                # again with the first meshgrid vector.
                # IndexError is raised in expressions like x[x > 0] since
                # "x > 0" evaluates to 'True', i.e. 1, and that index is
                # out of range for a meshgrid tuple of length 1 :-). To get
                # the real errors with indexing, we check again for the
                # same scenario (scalar output when not valid) as in the
                # first case.
                out = func_oop(x[0], **kwargs)

        else:
            # Here we don't catch exceptions since they are likely true
            # errors
            out = func_oop(x, **kwargs)

        def _process_array(out):
            if isinstance(out, backend.array_type) or np.isscalar(out):
                # Cast to proper dtype if needed, also convert to array if out  is a scalar.
                out = backend.array_constructor(
                    out,
                    dtype=backend_scalar_out_dtype,
                    device=device
                    )
                if scalar_in:
                    out = array_ns.squeeze(out,0)
                elif ndim == 1 and out.shape == (1,) + out_shape:
                    out = out.reshape(out_shape)

                if out_shape != () and out.shape != out_shape:
                    # Broadcast the returned element, but not in the
                    # scalar case. The resulting array may be read-only,
                    # in which case we copy.
                    out = array_ns.broadcast_to(out, out_shape)
                    out = backend.array_constructor(out, copy=True)
                return out

            elif isinstance(out, (tuple, list)):
                result = []
                assert len(out) != 0
                for sub_out in out:
                    result.append(_process_array(sub_out))
                return result

        return _process_array(out)

    return dual_use_func


if __name__ == '__main__':
    from odl.core.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2020 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Lebesgue L^p type discretizations of function spaces."""

from __future__ import absolute_import, division, print_function

from numbers import Integral

from contextlib import contextmanager

import numpy as np

from odl.core.discr.discr_utils import point_collocation, sampling_function
from odl.core.discr.partition import (
    RectPartition, uniform_partition, uniform_partition_fromintv)
from odl.core.set import IntervalProd, RealNumbers
from odl.core.set.space import LinearSpace, SupportedNumOperationParadigms, NumOperationParadigmSupport
from odl.core.space import ProductSpace
from odl.core.space.base_tensors import Tensor, TensorSpace, default_dtype
from odl.core.space.entry_points import tensor_space_impl
from odl.core.space.weightings.weighting import ConstWeighting
from odl.core.util import (
    apply_on_boundary, array_str, dtype_str, is_floating_dtype,
    is_numeric_dtype, normalized_nodes_on_bdry, normalized_scalar_param_list,
    repr_string, safe_int_conv, signature_string_parts)

__all__ = (
    'DiscretizedSpace',
    'DiscretizedSpaceElement',
    'uniform_discr_frompartition',
    'uniform_discr_fromintv',
    'uniform_discr',
    'uniform_discr_fromdiscr',
)


class DiscretizedSpace(TensorSpace):

    """Discretization of a Lebesgue :math:`L^p` space."""

    def __init__(self, partition, tspace, **kwargs):
        """Initialize a new instance.

        Parameters
        ----------
        partition : `RectPartition`
            Partition of a rectangular spatial domain.
        tspace : `TensorSpace`
            Space of elements used for data storage. It must have the same
            `TensorSpace.shape` as ``partition``.
        axis_labels : sequence of str, optional
            Names of the axes to use for plotting etc.
            Default:

            - 1D: ``['$x$']``
            - 2D: ``['$x$', '$y$']``
            - 3D: ``['$x$', '$y$', '$z$']``
            - nD: ``['$x_1$', '$x_2$', ..., '$x_n$']``

            Note: The ``$`` signs ensure rendering as LaTeX.
        """
        if not isinstance(partition, RectPartition):
            raise TypeError('`partition` must be a `RectPartition`, got {!r}'
                            ''.format(partition))
        if not isinstance(tspace, TensorSpace):
            raise TypeError('`tspace` must be a `TensorSpace`, got {!r}'
                            ''.format(tspace))
        if partition.shape != tspace.shape:
            raise ValueError(
                '`partition.shape` must be equal to `tspace.shape`, but '
                '{} != {}'.format(partition.shape, tspace.shape)
            )

        self.__tspace = tspace
        self.__partition = partition

        self._init_dtype(tspace.dtype)

        self._init_shape(tspace.shape, tspace.dtype)

        self._init_device(tspace.device)

        self.__use_in_place_ops = kwargs.pop('use_in_place_ops', True)
  
        self._init_weighting()

        field = self._init_field()

        LinearSpace.__init__(self, field)

        # Set axis labels
        axis_labels = kwargs.pop('axis_labels', None)
        if axis_labels is None:
            if self.ndim <= 3:
                self.__axis_labels = ('$x$', '$y$', '$z$')[:self.ndim]
            else:
                self.__axis_labels = tuple('$x_{}$'.format(axis)
                                           for axis in range(self.ndim))
        else:
            self.__axis_labels = tuple(str(label) for label in axis_labels)

        if kwargs:
            raise ValueError('got unexpected keyword arguments {}'
                             ''.format(kwargs))

    # --- Meta-info

    @property
    def element_type(self):
        """`DiscretizedSpaceElement`"""
        return DiscretizedSpaceElement

    @property
    def supported_num_operation_paradigms(self) -> NumOperationParadigmSupport:
        """In-place vs out-of-place is not of much concern for the discretization
        and only depends on the underlying arrays."""
        return self.tspace.supported_num_operation_paradigms

    # --- Constructor args

    @property
    def partition(self):
        """`RectPartition` of the function domain."""
        return self.__partition

    @property
    def tspace(self):
        """Space for the coefficients of the elements of this space."""
        return self.__tspace

    @property
    def axis_labels(self):
        """Labels for axes when displaying space elements."""
        return self.__axis_labels

    # --- Pass-through `partition` attributes

    @property
    def domain(self):
        """Set on which functions are defined before discretization."""
        return self.partition.set

    # --- Pass-through `tspace` attributes

    @property
    def weighting(self):
        """This space's weighting scheme."""
        # TODO(kohr-h): `weighting` is optional in `tspace`, how should we
        # handle that?
        return self.tspace.weighting

    @property
    def is_weighted(self):
        """``True`` if the ``tspace`` is weighted."""
        return getattr(self.tspace, 'is_weighted', False)

    @property
    def impl(self):
        """Name of the implementation back-end."""
        return self.tspace.impl

    @property
    def exponent(self):
        """Exponent of this space, the ``p`` in ``L^p``."""
        # TODO(kohr-h): `exponent` is optional in `tspace`, how should we
        # handle that?
        return self.tspace.exponent

    @property
    def min_pt(self):
        """Vector of minimal coordinates of the function domain."""
        return self.partition.min_pt

    @property
    def max_pt(self):
        """Vector of maximal coordinates of the function domain."""
        return self.partition.max_pt

    @property
    def is_uniform_byaxis(self):
        """Boolean tuple showing uniformity of ``self.partition`` per axis."""
        return self.partition.is_uniform_byaxis

    @property
    def is_uniform(self):
        """``True`` if `partition` is uniform."""
        return self.partition.is_uniform

    @property
    def grid(self):
        """Sampling grid of the discretization mappings."""
        return self.partition.grid

    @property
    def shape(self):
        """Shape of the underlying partition."""
        return self.partition.shape

    @property
    def ndim(self):
        """Number of dimensions (= number of axes)."""
        return self.partition.ndim

    @property
    def size(self):
        """Total number of underlying partition cells."""
        return self.partition.size

    @property
    def cell_sides(self):
        """Side lengths of a cell in an underlying *uniform* partition."""
        return self.partition.cell_sides

    @property
    def cell_volume(self):
        """Cell volume of an underlying *uniform* partition."""
        return self.partition.cell_volume

    @property
    def meshgrid(self):
        """All sampling points in the partition as a sparse meshgrid."""
        return self.partition.meshgrid

    def points(self):
        """All sampling points in the partition.

        Returns
        -------
        points : `numpy.ndarray`
            The shape of the array is ``size x ndim``, i.e. the points
            are stored as rows.
        """
        return self.partition.points()

    def available_dtypes(self):
        """Available data types for new elements in this space.

        This is equal to the available data types of `tspace`.
        """
        return self.tspace.available_dtypes()

    # --- Derived properties

    @property
    def tspace_type(self):
        """Tensor space type of this space."""
        return type(self.tspace)

    @property
    def tangent_bundle(self):
        """The tangent bundle associated with `domain` using `partition`.

        The tangent bundle of a space ``X`` of functions ``R^d --> F`` can be
        interpreted as the space of vector-valued functions ``R^d --> F^d``.
        This space can be identified with the power space ``X^d`` as used
        in this implementation.
        """
        if self.ndim == 0:
            return ProductSpace(field=self.field)
        else:
            return ProductSpace(self, self.ndim)

    @property
    def is_uniformly_weighted(self):
        """``True`` if the weighting is the same for all space points."""
        try:
            is_uniformly_weighted = self.__is_uniformly_weighted
        except AttributeError:
            bdry_fracs = self.partition.boundary_cell_fractions
            is_uniformly_weighted = (
                np.allclose(bdry_fracs, 1.0) or
                self.exponent == float('inf') or
                not getattr(self.tspace, 'is_weighted', False))

            self.__is_uniformly_weighted = is_uniformly_weighted

        return is_uniformly_weighted

    # --- Element creation

    def element(self, inp=None, **kwargs):
        """Create an element from ``inp`` or from scratch.

        Parameters
        ----------
        inp : optional
            Input used to initialize the new element. The following options
            are available:

            - ``None``: an empty element is created with no guarantee of
              its state (memory allocation only).

            - array-like: an element wrapping a `tensor` is created,
              where a copy is avoided whenever possible. This usually
              requires correct `shape`, `dtype` and `impl` if applicable.
              See the ``element`` method of `tspace` for more
              information.

              If any of these conditions is not met, a copy is made.

            - callable: a new element is created by sampling the function
              using `point_collocation`.

        kwargs :
            Additional arguments passed on to `point_collocation` when
            called on ``inp``, in the form
            ``point_collocation(inp, points, **kwargs)``.
            This can be used e.g. for functions with parameters.

        Returns
        -------
        element : `DiscretizedSpaceElement`
            The discretized element, calculated as ``point_collocation(inp)``
            or ``tspace.element(inp)``, tried in this order.

        Examples
        --------
        Elements can be created from array-like objects that represent
        an already discretized function:

        >>> space = odl.uniform_discr(-1, 1, 4)
        >>> space.element([1, 2, 3, 4])
        uniform_discr(-1.0, 1.0, 4).element([ 1.,  2.,  3.,  4.])
        >>> vector = odl.rn(4).element([0, 1, 2, 3])
        >>> space.element(vector)
        uniform_discr(-1.0, 1.0, 4).element([ 0.,  1.,  2.,  3.])

        On the other hand, non-discretized objects like Python functions
        can be discretized "on the fly":

        >>> space.element(lambda x: x * 2)
        uniform_discr(-1.0, 1.0, 4).element([-1.5, -0.5, 0.5, 1.5])

        This works also with parameterized functions, however only
        through keyword arguments (not positional arguments with
        defaults):

        >>> def f(x, c=0.0):
        ...     return np.maximum(x, c)
        ...
        >>> space = odl.uniform_discr(-1, 1, 4)
        >>> space.element(f, c=0.5)
        uniform_discr(-1.0, 1.0, 4).element([ 0.5 ,  0.5 ,  0.5 ,  0.75])
        """
        if 'order' in kwargs:
            raise RuntimeError('The use of the order argument is now deprecated, please remove it. All arrays are C contiguous.')
        if inp is None:
            return self.element_type(self, self.tspace.element())
        elif inp in self:
            return inp
        elif inp in self.tspace:
            return self.element_type(self, inp)
        elif callable(inp):
            func = sampling_function(
                inp, self.domain, out_dtype=self.dtype_identifier, impl=self.impl, device=self.device
            )
            sampled = point_collocation(func, self.meshgrid, **kwargs)
            return self.element_type(
                self, self.tspace.element(sampled)
            )
        else:
            # Sequence-type input
            return self.element_type(
                self, self.tspace.element(inp)
            )

    def zero(self):
        """Return the element of all zeros."""
        return self.element_type(self, self.tspace.zero())

    def one(self):
        """Return the element of all ones."""
        return self.element_type(self, self.tspace.one())

    # --- Casting

    def _astype(self, dtype):
        """Internal helper for ``astype``."""
        tspace = self.tspace.astype(dtype)
        return type(self)(
            self.partition, tspace, axis_labels=self.axis_labels)

    def _to_device(self, device:str):
        """Internal helper for `to_device`."""
        tspace = self.tspace.to_device(device)
        return type(self)(
            self.partition, tspace, axis_labels=self.axis_labels)
    
    def _to_impl(self, impl:str):
        """Internal helper for `to_impl`."""
        tspace = self.tspace.to_impl(impl)
        return type(self)(
            self.partition, tspace, axis_labels=self.axis_labels)
    
    # --- Slicing

    # TODO: add `byaxis`_out when discretized tensor-valued functions are
    # available

    @property
    def byaxis_in(self):
        """Object to index along input (domain) dimensions.

        Examples
        --------
        Indexing with integers or slices:

        >>> space = odl.uniform_discr([0, 0, 0], [1, 2, 3], (5, 10, 15))
        >>> space.byaxis_in[0]
        uniform_discr(0.0, 1.0, 5)
        >>> space.byaxis_in[1]
        uniform_discr(0.0, 2.0, 10)
        >>> space.byaxis_in[1:]
        uniform_discr([ 0.,  0.], [ 2.,  3.], (10, 15))

        Lists can be used to stack spaces arbitrarily:

        >>> space.byaxis_in[[2, 1, 2]]
        uniform_discr([ 0.,  0.,  0.], [ 3.,  2.,  3.], (15, 10, 15))
        """
        space = self

        class DiscretizedSpaceByaxisIn(object):

            """Helper class for indexing by domain axes."""

            def __getitem__(self, indices):
                """Return ``self[indices]``.

                Parameters
                ----------
                indices : index expression
                    Object used to index the space domain.

                Returns
                -------
                space : `DiscretizedSpace`
                    The resulting space with indexed domain and otherwise
                    same properties (except possibly weighting).
                """
                part = space.partition.byaxis[indices]

                if isinstance(space.weighting, ConstWeighting):
                    # Need to manually construct `tspace` since it doesn't
                    # know where its weighting factor comes from
                    try:
                        iter(indices)
                    except TypeError:
                        newshape = space.shape[indices]
                    else:
                        newshape = tuple(space.shape[int(i)] for i in indices)

                    weighting = part.cell_volume
                    tspace = type(space.tspace)(
                        newshape, space.dtype,
                        exponent=space.exponent, weighting=weighting)
                else:
                    # Other weighting schemes are handled correctly by
                    # the tensor space
                    # TODO(kohr-h): `byaxis` is not guaranteed to exist in
                    # `tspace`, how to handle that?
                    tspace = space.tspace.byaxis[indices]

                try:
                    iter(indices)
                except TypeError:
                    labels = space.axis_labels[indices]
                else:
                    labels = tuple(space.axis_labels[int(i)]
                                   for i in indices)

                return DiscretizedSpace(part, tspace, axis_labels=labels)

            def __repr__(self):
                """Return ``repr(self)``."""
                return repr(space) + '.byaxis_in'

        return DiscretizedSpaceByaxisIn()

    # --- Identity

    def __eq__(self, other):
        """Return ``self == other``.

        Returns
        -------
        equals : bool
            ``True`` if ``other`` is a `DiscretizedSpace` with equal
            `tspace`, ``False`` otherwise.
        """
        # Optimizations for simple cases
        if other is self:
            return True
        elif other is None:
            return False
        else:
            return (
                super(DiscretizedSpace, self).__eq__(other)
                and other.tspace == self.tspace
                and other.partition == self.partition
            )

    def __hash__(self):
        """Return ``hash(self)``."""
        return hash(
            (super(DiscretizedSpace, self).__hash__(),
             self.tspace,
             self.partition)
        )

    # --- Space functions

    def _lincomb(self, a, x1, b, x2, out):
        """Raw linear combination."""
        return self.element(
            self.tspace._lincomb(a, x1.tensor, b, x2.tensor,
                                 out.tensor if out is not None else None))

    def _multiply(self, x1, x2, out):
        """Raw pointwise multiplication of two elements."""
        return self.element(
                self.tspace._multiply(x1.tensor, x2.tensor,
                                      out.tensor if out is not None else None))

    def _divide(self, x1, x2, out):
        """Raw pointwise multiplication of two elements."""
        return self.element(
                self.tspace._divide(x1.tensor, x2.tensor,
                                    out.tensor if out is not None else None))

    # The inherited methods by default use a weighting by a constant
    # (the grid cell size). In dimensions where the partitioned set contains
    # only a fraction of the outermost cells (e.g. if the outermost grid
    # points lie at the boundary), the corresponding contributions to
    # discretized integrals need to be scaled by that fraction.
    def _inner(self, x, y):
        """Return ``self.inner(x, y)``."""
        if self.is_uniform and not self.is_uniformly_weighted:
            # TODO: implement without copying x
            bdry_fracs = self.partition.boundary_cell_fractions
            func_list = _scaling_func_list(bdry_fracs, exponent=1.0)
            x_arr = apply_on_boundary(x, func=func_list, only_once=False)
            return self.tspace.inner(self.tspace.element(x_arr), y.tensor)
        else:
            return self.tspace.inner(x.tensor, y.tensor)

    def _norm(self, x):
        """Return ``self.norm(x)``."""
        if self.is_uniform and not self.is_uniformly_weighted:
            # TODO: implement without copying x
            bdry_fracs = self.partition.boundary_cell_fractions
            func_list = _scaling_func_list(bdry_fracs, exponent=self.exponent)
            x_arr = apply_on_boundary(x, func=func_list, only_once=False)
            return self.tspace.norm(self.tspace.element(x_arr))
        else:
            return self.tspace.norm(x.tensor)

    def _dist(self, x, y):
        """Return ``self.dist(x, y)``."""
        if self.is_uniform and not self.is_uniformly_weighted:
            bdry_fracs = self.partition.boundary_cell_fractions
            func_list = _scaling_func_list(bdry_fracs, exponent=self.exponent)
            arrs = [apply_on_boundary(vec, func=func_list, only_once=False)
                    for vec in (x, y)]

            return self.tspace.dist(
                self.tspace.element(arrs[0]),
                self.tspace.element(arrs[1]),
            )
        else:
            return self.tspace.dist(x.tensor, y.tensor)

    def __repr__(self):
        """Return ``repr(self)``."""
        # Clunky check if the factory repr can be used
        if uniform_partition_fromintv(
            self.partition.set, self.shape, nodes_on_bdry=False
        ) == self.partition:
            use_uniform = True
            nodes_on_bdry = False
        elif uniform_partition_fromintv(
            self.partition.set, self.shape, nodes_on_bdry=True
        ) == self.partition:
            use_uniform = True
            nodes_on_bdry = True
        else:
            use_uniform = False
            nodes_on_bdry = None

        if use_uniform:
            ctor = 'uniform_discr'
            if self.ndim == 1:
                posargs = [self.min_pt[0], self.max_pt[0], self.shape[0]]
                posmod = ['', '', '']
            else:
                posargs = [self.min_pt, self.max_pt, self.shape]
                posmod = [array_str, array_str, '']

            default_dtype_s = dtype_str(
                default_dtype(self.tspace.array_backend, RealNumbers())
            )

            dtype_s = dtype_str(self.dtype)
            optargs = [
                ('impl', self.impl, 'numpy'),
                ('nodes_on_bdry', nodes_on_bdry, False),
                ('dtype', dtype_s, default_dtype_s)
            ]

            # Add weighting stuff if not equal to default
            if (
                self.exponent == float('inf')
                or self.ndim == 0
                or not is_floating_dtype(self.dtype)
            ):
                # In these cases, weighting constant 1 is the default
                if (
                    not isinstance(self.weighting, ConstWeighting)
                    or not np.isclose(self.weighting.const, 1.0)
                ):
                    optargs.append(('weighting', self.weighting.const, None))
            else:
                if (
                    not isinstance(self.weighting, ConstWeighting)
                    or not np.isclose(self.weighting.const, self.cell_volume)
                ):
                    optargs.append(('weighting', self.weighting.const, None))

            optmod = [''] * len(optargs)

            # The following applies only if a shorthand Python-type version of
            # the dtype is shown. In that case, no quotation marks should be put
            # around it. This is not enabled because `dtype_str` currently always
            # produces a conventional string (e.g. `'complex128'` instead of
            # `complex`.
            if False and self.dtype in (float, complex, int, bool):
                optmod[2] = '!s'

            inner_parts = signature_string_parts(
                posargs, optargs, [posmod, optmod]
            )
            return repr_string(ctor, inner_parts)

        else:
            ctor = self.__class__.__name__
            posargs = [self.partition, self.tspace]
            inner_parts = signature_string_parts(posargs, [])
            return repr_string(ctor, inner_parts, allow_mixed_seps=False)
        
    def __str__(self):
        """Return ``str(self)``."""
        return repr(self)


class DiscretizedSpaceElement(Tensor):

    """Representation of a `DiscretizedSpace` element."""

    def __init__(self, space, tensor):
        """Initialize a new instance."""
        super(DiscretizedSpaceElement, self).__init__(space)
        self.__tensor = tensor

    # --- Constructor args

    @property
    def tensor(self):
        """Structure for data storage."""
        return self.__tensor

    # --- Pass-through `space` properties

    @property
    def cell_sides(self):
        """Side lengths of a cell in an underlying *uniform* partition."""
        return self.space.cell_sides

    @property
    def cell_volume(self):
        """Cell volume of an underlying regular grid."""
        return self.space.cell_volume

    # --- Pass-through `tensor` properties

    @property
    def data(self):
        """Data container of ``self``, depends on ``space.impl``."""
        return self.tensor.data

    @property
    def dtype(self):
        """Type of data storage."""
        return self.tensor.dtype

    @property
    def size(self):
        """Size of data storage."""
        return self.tensor.size

    def __len__(self):
        """Return ``len(self)``.

        Equivalent to ``self.shape[0]`` if possible. Zero-dimensional
        tensors have no length and produce a `TypeError`.
        """
        return len(self.tensor)

    def copy(self):
        """Create an identical (deep) copy of this element."""
        return self.space.element(self.tensor.copy())

    def asarray(self, out=None, must_be_contiguous=False):
        """Extract the data of this array as a numpy array.

        Parameters
        ----------
        out : `numpy.ndarray`, optional
            Array in which the result should be written in-place.
            Has to be contiguous and of the correct dtype.
        """
        return self.tensor.asarray(out=out, must_be_contiguous=must_be_contiguous)

    @contextmanager
    def writable_array(self, must_be_contiguous: bool =False):
        arr = None
        try:
            arr = self.tensor.asarray(must_be_contiguous=must_be_contiguous)
            yield arr
        finally:
            if arr is not None:
                self.tensor.data[:] = arr

    def astype(self, dtype):
        """Return a copy of this element with new ``dtype``.

        Parameters
        ----------
        dtype :
            Scalar data type of the returned space. Can be provided
            in any way the `numpy.dtype` constructor understands, e.g.
            as built-in type or as a string. Data types with non-trivial
            shapes are not allowed.

        Returns
        -------
        newelem : `DisceteLpElement`
            Version of this element with given data type.
        """
        return self.space.astype(dtype).element(self.tensor.astype(dtype))

    def _assign(self, other, avoid_deep_copy):
        """Assign the values of ``other``, which is assumed to be in the
        same discretized space, to ``self``."""
        self.__tensor.assign(other.tensor, avoid_deep_copy=avoid_deep_copy)

    def __eq__(self, other):
        """Return ``self == other``.

        Returns
        -------
        equals : bool
            ``True`` if all entries of ``other`` are equal to this
            element's entries, ``False`` otherwise.
        """
        return other in self.space and other.tensor == self.tensor

    def __getitem__(self, indices):
        """Return ``self[indices]``.

        Parameters
        ----------
        indices : int or `slice`
            The position(s) that should be accessed.

        Returns
        -------
        values : `Tensor`
            The value(s) at the index (indices).
        """
        if isinstance(indices, type(self)):
            indices = indices.tensor
        return self.tensor[indices]

    def __ipow__(self, p):
        """Implement ``self **= p``."""
        # The concrete `tensor` can specialize `__ipow__` for non-integer
        # `p` so we want to use it here. Otherwise we get the default
        # `LinearSpaceElement.__ipow__` which only works for integer `p`.
        self.tensor.__ipow__(p)
        return self

    @property
    def real(self):
        """Real part of this element.

        Returns
        -------
        real : `DiscretizedSpaceElement`

        Examples
        --------
        Get the real part:

        >>> discr = odl.uniform_discr(0, 1, 3, dtype=complex)
        >>> x = discr.element([5+1j, 3, 2-2j])
        >>> x.real
        uniform_discr(0.0, 1.0, 3).element([ 5.,  3.,  2.])

        Set the real part:

        >>> x = discr.element([1 + 1j, 2, 3 - 3j])
        >>> zero = discr.real_space.zero()
        >>> x.real = zero
        >>> x.real
        uniform_discr(0.0, 1.0, 3).element([ 0.,  0.,  0.])

        Other array-like types and broadcasting:

        >>> x.real = 1.0
        >>> x.real
        uniform_discr(0.0, 1.0, 3).element([ 1.,  1.,  1.])
        >>> x.real = [2, 3, 4]
        >>> x.real
        uniform_discr(0.0, 1.0, 3).element([ 2.,  3.,  4.])
        """
        return self.space.real_space.element(self.tensor.real)

    @real.setter
    def real(self, newreal):
        """Set the real part of this element to ``newreal``.

        This method is invoked by ``x.real = other``.

        Parameters
        ----------
        newreal : array-like or scalar
            Values to be assigned to the real part of this element.
        """
        if isinstance(newreal, DiscretizedSpaceElement):
            self.tensor.real = newreal.tensor
        else:
            self.tensor.real = newreal

    @property
    def imag(self):
        """Imaginary part of this element.

        Returns
        -------
        imag : `DiscretizedSpaceElement`

        Examples
        --------
        Get the imaginary part:

        >>> discr = uniform_discr(0, 1, 3, dtype=complex)
        >>> x = discr.element([5+1j, 3, 2-2j])
        >>> x.imag
        uniform_discr(0.0, 1.0, 3).element([ 1.,  0., -2.])

        Set the imaginary part:

        >>> x = discr.element([1 + 1j, 2, 3 - 3j])
        >>> zero = discr.real_space.zero()
        >>> x.imag = zero
        >>> x.imag
        uniform_discr(0.0, 1.0, 3).element([ 0.,  0.,  0.])

        Other array-like types and broadcasting:

        >>> x.imag = 1.0
        >>> x.imag
        uniform_discr(0.0, 1.0, 3).element([ 1.,  1.,  1.])
        >>> x.imag = [2, 3, 4]
        >>> x.imag
        uniform_discr(0.0, 1.0, 3).element([ 2.,  3.,  4.])
        """
        return self.space.real_space.element(self.tensor.imag)

    @imag.setter
    def imag(self, newimag):
        """Set the imaginary part of this element to ``newimag``.

        This method is invoked by ``x.imag = other``.

        Parameters
        ----------
        newimag : array-like or scalar
            Values to be assigned to the imaginary part of this element.

        Raises
        ------
        ValueError
            If the space is real, i.e., no imagninary part can be set.
        """
        if self.space.is_real:
            raise ValueError('cannot set imaginary part in real spaces')
        if isinstance(newimag, DiscretizedSpaceElement):
            self.tensor.imag = newimag.tensor
        else:
            self.tensor.imag = newimag

    def conj(self, out=None):
        """Complex conjugate of this element.

        Parameters
        ----------
        out : `DiscretizedSpaceElement`, optional
            Element to which the complex conjugate is written.
            Must be an element of this element's space.

        Returns
        -------
        out : `DiscretizedSpaceElement`
            The complex conjugate element. If ``out`` is provided,
            the returned object is a reference to it.

        Examples
        --------
        >>> discr = uniform_discr(0, 1, 4, dtype=complex)
        >>> x = discr.element([5+1j, 3, 2-2j, 1j])
        >>> y = x.conj()
        >>> print(y)
        [ 5.-1.j,  3.-0.j,  2.+2.j,  0.-1.j]

        The out parameter allows you to avoid a copy:

        >>> z = discr.element()
        >>> z_out = x.conj(out=z)
        >>> print(z)
        [ 5.-1.j,  3.-0.j,  2.+2.j,  0.-1.j]
        >>> z_out is z
        True

        It can also be used for in-place conjugation:

        >>> x_out = x.conj(out=x)
        >>> print(x)
        [ 5.-1.j,  3.-0.j,  2.+2.j,  0.-1.j]
        >>> x_out is x
        True
        """
        if out is None:
            return self.space.element(self.tensor.conj())
        else:
            self.tensor.conj(out=out.tensor)
            return out

    def __setitem__(self, indices, values):
        """Set values of this element.

        Parameters
        ----------
        indices : int or `slice`
            The position(s) that should be set
        values : scalar or `array-like`
            Value(s) to be assigned.
            If ``indices`` is an integer, ``values`` must be a scalar
            value.
            If ``indices`` is a slice, ``values`` must be
            broadcastable to the size of the slice (same size,
            shape ``(1,)`` or scalar).
            For ``indices == slice(None)``, i.e. in the call
            ``vec[:] = values``, a multi-dimensional array of correct
            shape is allowed as ``values``.
        """
        if values in self.space:
            self.tensor[indices] = values.tensor
        else:
            if isinstance(indices, type(self)):
                indices = indices.tensor
            if isinstance(values, type(self)):
                values = values.tensor
            self.tensor.__setitem__(indices, values)


    def show(self, title=None, method='', coords=None, indices=None,
             force_show=False, fig=None, **kwargs):
        """Display the function graphically.

        Parameters
        ----------
        title : string, optional
            Set the title of the figure

        method : string, optional
            1d methods:

            - ``'plot'`` : graph plot (default for 1d data)
            - ``'scatter'`` : scattered 2d points (2nd axis <-> value)

            2d methods:

            - ``'imshow'`` : image plot with coloring according to value,
              including a colorbar (default for 2d data).
            - ``'scatter'`` : cloud of scattered 3d points
              (3rd axis <-> value)

        coords : `array-like`, optional
            Display a slice of the array instead of the full array.
            The values are shown accordinging to the given values,
            where ``None`` means all values along that dimension. For
            example, ``[None, None, 0.5]`` shows all values in the first
            two dimensions, with the third coordinate equal to 0.5.
            If a sequence is provided, it specifies the minimum and maximum
            point to be shown, i.e. ``[None, [0, 1]]`` shows all of the
            first axis and values between 0 and 1 in the second.
            This option is mutually exclusive with ``indices``.

        indices : int, slice, Ellipsis or sequence, optional
            Display a slice of the array instead of the full array.
            If a sequence is given, the i-th entry indexes the i-th axis,
            with the following behavior for the different types of entries:

                - ``int``: take entries with this index along axis ``i``,
                  removing this axis from the result
                - ``slice``: take a subset along axis ``i``, keeping it
                  intact
                - ``None``: equivalent to ``slice(None)``
                - ``Ellipsis`` (``...``): equivalent to the number of
                  ``None`` entries required to fill up the sequence to
                  correct length.

            The typical use case is to show a slice for a fixed index in
            a specific axis, which can be done most easily by setting, e.g.,
            ``indices=[None, 50, None]`` to take the 2d slice parallel to
            the x-z coordinate plane at index ``y = 50``.

            A single ``int`` or ``slice`` object indexes the first
            axis, i.e., is treated as ``(int_or_slice, Ellipsis)``.
            For the default ``None``, the array is kepty as-is for data
            that has at most 2 dimensions. For higher-dimensional
            data, the 2d slice in the first two axes at the middle
            position along the remaining axes is shown
            (semantically ``[:, :, shape[2:] // 2]``).
            This option is mutually exclusive with ``coords``.

        force_show : bool, optional
            Whether the plot should be forced to be shown now or deferred until
            later. Note that some backends always displays the plot, regardless
            of this value.

        fig : `matplotlib.figure.Figure`, optional
            The figure to show in. Expected to be of same "style", as
            the figure given by this function. The most common use case
            is that ``fig`` is the return value of an earlier call to
            this function.

        Other Parameters
        ----------------
        interp : {'linear', 'nearest'}, optional
            Interpolation type that should be used for the plot.

        kwargs : {'figsize', 'saveto', 'clim', ...}, optional
            Extra keyword arguments passed on to the display method.
            See the Matplotlib functions for documentation of extra
            options.

        Returns
        -------
        fig : `matplotlib.figure.Figure`
            The resulting figure. It is also shown to the user.

        See Also
        --------
        odl.core.util.graphics.show_discrete_data : Underlying implementation
        """
        from odl.core.util.graphics import show_discrete_data

        if 'interp' not in kwargs:
            kwargs['interp'] = 'linear'

        if self.ndim == 0:
            raise ValueError('nothing to show for 0-dimensional vector')

        if coords is not None:
            if indices is not None:
                raise ValueError('cannot provide both coords and indices')

            partition = self.space.partition
            shape = self.shape
            indices = []
            for axis, (n, coord) in enumerate(zip(shape, coords)):
                try:
                    coord_minp, coord_maxp = coord
                except TypeError:
                    coord_minp = coord_maxp = coord

                subpart = partition.byaxis[axis]

                # Validate input
                if coord_minp is not None:
                    coord_minp = subpart.set.element(coord_minp)
                if coord_maxp is not None:
                    coord_maxp = subpart.set.element(coord_maxp)

                if len(subpart) == 0:  # trivial cases
                    indices.append(0)
                elif coord_minp is not None and coord_minp == coord_maxp:
                    indices.append(subpart.index(coord_minp))
                else:
                    if coord_minp is None:
                        min_ind = 0
                    else:
                        min_ind = np.floor(subpart.index(coord_minp,
                                                         floating=True))

                    if coord_maxp is None:
                        max_ind = len(subpart)
                    else:
                        max_ind = np.ceil(subpart.index(coord_maxp,
                                                        floating=True))

                    indices.append(slice(int(min_ind), int(max_ind)))

        # Default to showing x-y slice "in the middle"
        if indices is None and self.ndim >= 3:
            indices = ((slice(None),) * 2 +
                       tuple(n // 2 for n in self.space.shape[2:]))

        # Normalize indices
        if isinstance(indices, (Integral, slice)):
            indices = (indices,)
        elif indices is None or indices == Ellipsis:
            indices = (slice(None),) * self.ndim

        # Single index or slice indexes the first axis, rest untouched
        if len(indices) == 1:
            indices = tuple(indices) + (Ellipsis,)

        # Convert `Ellipsis` objects
        if indices.count(Ellipsis) > 1:
            raise ValueError('cannot use more than 1 `Ellipsis` (`...`)')
        elif Ellipsis in indices:
            # Replace Ellipsis with the correct number of `slice(None)`
            pos = indices.index(Ellipsis)
            indices = (tuple(indices[:pos]) +
                       (slice(None),) * (self.ndim - len(indices) + 1) +
                       tuple(indices[pos + 1:]))

        # Now indices should be exactly of length `ndim`
        if len(indices) < self.ndim:
            raise ValueError('too few axes ({} < {})'.format(len(indices),
                                                             self.ndim))
        if len(indices) > self.ndim:
            raise ValueError('too many axes ({} > {})'.format(len(indices),
                                                              self.ndim))

        # Map `None` to `slice(None)` in indices for syntax like `coords`
        indices = tuple(slice(None) if idx is None else idx
                        for idx in indices)

        squeezed_axes = [axis for axis in range(self.ndim)
                         if not isinstance(indices[axis], Integral)]
        axis_labels = [self.space.axis_labels[axis] for axis in squeezed_axes]

        # Squeeze grid and values according to the index expression
        part = self.space.partition[indices].squeeze()
        values = self.asarray()[indices].squeeze()

        return show_discrete_data(values, part, title=title, method=method,
                                  force_show=force_show, fig=fig,
                                  axis_labels=axis_labels, **kwargs)


def uniform_discr_frompartition(partition, dtype=None, impl='numpy', **kwargs):
    """Return a uniformly discretized L^p function space.

    Parameters
    ----------
    partition : `RectPartition`
        Uniform partition to be used for discretization.
        It defines the domain and the functions and the grid for
        discretization.
    dtype : optional
        Data type for the discretized space, must be understood by the
        `numpy.dtype` constructor. The default for ``None`` depends on the
        ``impl`` backend, usually it is ``'float64'`` or ``'float32'``.
    impl : string, optional
        Implementation of the data storage arrays
    kwargs :
        Additional keyword parameters, see `uniform_discr` for details.

    Returns
    -------
    discr : `DiscretizedSpace`
        The uniformly discretized function space.

    Examples
    --------
    >>> part = odl.uniform_partition(0, 1, 10)
    >>> uniform_discr_frompartition(part)
    uniform_discr(0.0, 1.0, 10)

    See Also
    --------
    uniform_discr : implicit uniform Lp discretization
    uniform_discr_fromspace : uniform Lp discretization from an existing
        function space
    odl.core.discr.partition.uniform_partition :
        partition of the function domain
    """
    if not isinstance(partition, RectPartition):
        raise TypeError('`partition` {!r} is not a `RectPartition` instance'
                        ''.format(partition))
    if not partition.is_uniform:
        raise ValueError('`partition` is not uniform')

    # if dtype is not None:
    #     dtype = np.dtype(dtype)

    tspace_type = tensor_space_impl(impl)
    if dtype is None:
        dtype = default_dtype(impl)

    weighting = kwargs.pop('weighting', None)
    exponent = kwargs.pop('exponent', 2.0)
    if weighting is None and is_numeric_dtype(dtype):
        if exponent == float('inf') or partition.ndim == 0:
            weighting = 1.0
        else:
            weighting = partition.cell_volume

    device = kwargs.pop('device', 'cpu')
    tspace = tspace_type(partition.shape, dtype, exponent=exponent,
                         weighting=weighting, device=device)
    return DiscretizedSpace(partition, tspace, **kwargs)


def uniform_discr_fromintv(intv_prod, shape, dtype=None, impl='numpy',
                           **kwargs):
    """Return a uniformly discretized L^p function space.

    Parameters
    ----------
    intv_prod : `IntervalProd`
        Function domain of the uniformly discretized space.
    shape : int or sequence of ints
        Number of samples per axis.
    dtype : optional
        Data type for the discretized space, must be understood by the
        `numpy.dtype` constructor. The default for ``None`` depends on the
        ``impl`` backend, usually it is ``'float64'`` or ``'float32'``.
    impl : str, optional
        Implementation of the data storage arrays.
    kwargs :
        Additional keyword parameters, see `uniform_discr` for details.

    Returns
    -------
    discr : `DiscretizedSpace`
        The uniformly discretized function space

    Examples
    --------
    >>> intv = IntervalProd(0, 1)
    >>> uniform_discr_fromintv(intv, 10)
    uniform_discr(0.0, 1.0, 10)

    See Also
    --------
    uniform_discr : implicit uniform Lp discretization
    uniform_discr_frompartition : uniform Lp discretization using a given
        uniform partition of a function domain
    """
    if dtype is None:
        dtype = default_dtype(impl)

    nodes_on_bdry = kwargs.pop('nodes_on_bdry', False)
    partition = uniform_partition_fromintv(intv_prod, shape, nodes_on_bdry)
    return uniform_discr_frompartition(partition, dtype, impl, **kwargs)


def uniform_discr(min_pt, max_pt, shape, dtype=None, impl='numpy', **kwargs):
    """Return a uniformly discretized L^p function space.

    Parameters
    ----------
    min_pt, max_pt : float or sequence of floats
        Minimum/maximum corners of the desired function domain.
    shape : int or sequence of ints
        Number of samples per axis.
    dtype : optional
        Data type for the discretized space, must be understood by the
        `numpy.dtype` constructor. The default for ``None`` depends on the
        ``impl`` backend, usually it is ``'float64'`` or ``'float32'``.
    impl : string, optional
        Implementation of the data storage arrays.

    Other Parameters
    ----------------
    exponent : positive float, optional
        The parameter :math:`p` in :math:`L^p`. If the exponent is not
        equal to the default 2.0, the space has no inner product.
    nodes_on_bdry : bool or sequence, optional
        If a sequence is provided, it determines per axis whether to
        place the last grid point on the boundary (``True``) or shift it
        by half a cell size into the interior (``False``). In each axis,
        an entry may consist in a single bool or a 2-tuple of
        bool. In the latter case, the first tuple entry decides for
        the left, the second for the right boundary. The length of the
        sequence must be ``len(shape)``.
        A single boolean is interpreted as a global choice for all
        boundaries.
        Default: ``False``.
    weighting : optional
        Use weighted inner product, norm, and dist. The following
        types are supported as ``weighting``:

        - ``None``: Use the cell volume as weighting constant (default).
        - ``float``: Weighting by a constant.
        - array-like: Point-wise weighting by an array.
        - `Weighting`: Use weighting class as-is. Compatibility
          with this space's elements is not checked during init.

    Returns
    -------
    discr : `DiscretizedSpace`
        The uniformly discretized function space

    Examples
    --------
    Create real space:

    >>> space = uniform_discr([0, 0], [1, 1], (10, 10))
    >>> space
    uniform_discr([ 0.,  0.], [ 1.,  1.], (10, 10))
    >>> space.cell_sides
    array([ 0.1,  0.1])
    >>> space.dtype
    dtype('float64')
    >>> space.is_real
    True

    Create complex space by giving a dtype:

    >>> space = uniform_discr([0, 0], [1, 1], (10, 10), dtype=complex)
    >>> space
    uniform_discr([ 0.,  0.], [ 1.,  1.], (10, 10), dtype='complex128')
    >>> space.is_complex
    True
    >>> space.real_space  # Get real counterpart
    uniform_discr([ 0.,  0.], [ 1.,  1.], (10, 10))

    See Also
    --------
    uniform_discr_frompartition : uniform Lp discretization using a given
        uniform partition of a function domain
    uniform_discr_fromspace : uniform discretization from an existing
        function space
    uniform_discr_fromintv : uniform discretization from an existing
        interval product
    """
    intv_prod = IntervalProd(min_pt, max_pt)
    return uniform_discr_fromintv(intv_prod, shape, dtype, impl, **kwargs)


def uniform_discr_fromdiscr(discr, min_pt=None, max_pt=None,
                            shape=None, cell_sides=None, **kwargs):
    """Return a discretization based on an existing one.

    The parameters that are explicitly given are used to create the
    new discretization, and the missing information is taken from
    the template space. See Notes for the exact functionality.

    Parameters
    ----------
    discr : `DiscretizedSpace`
        Uniformly discretized space used as a template.
    min_pt, max_pt: float or sequence of floats, optional
        Desired minimum/maximum corners of the new space domain.
    shape : int or sequence of ints, optional
        Desired number of samples per axis of the new space.
    cell_sides : float or sequence of floats, optional
        Desired cell side lenghts of the new space's partition.

    Other Parameters
    ----------------
    nodes_on_bdry : bool or sequence, optional
        If a sequence is provided, it determines per axis whether to
        place the last grid point on the boundary (``True``) or shift it
        by half a cell size into the interior (``False``). In each axis,
        an entry may consist in a single bool or a 2-tuple of
        bool. In the latter case, the first tuple entry decides for
        the left, the second for the right boundary. The length of the
        sequence must be ``discr.ndim``.

        A single boolean is interpreted as a global choice for all
        boundaries.

        Default: ``False``.

    kwargs :
        Additional keyword parameters passed to the `DiscretizedSpace`
        initializer.

    Notes
    -----
    The parameters ``min_pt``, ``max_pt``, ``shape`` and
    ``cell_sides`` can be combined in the following ways (applies in
    each axis individually):

    **0 arguments:**
        Return a copy of ``discr``

    **1 argument:**
        [min,max]_pt -> keep sampling but translate domain so it
        starts/ends at ``[min,max]_pt``

        shape/cell_sides -> keep domain but change sampling.
        See `uniform_partition` for restrictions.

    **2 arguments:**
        min_pt + max_pt -> translate and resample with the same
        number of samples

        [min,max]_pt + shape/cell_sides -> translate and resample

        shape + cell_sides -> error due to ambiguity (keep
        ``min_pt`` or ``max_pt``?)

    **3+ arguments:**
        The underlying partition is uniquely determined by the new
        parameters. See `uniform_partition`.

    See Also
    --------
    uniform_discr : implicit uniform Lp discretization
    odl.core.discr.partition.uniform_partition :
        underlying domain partitioning scheme

    Examples
    --------
    >>> discr = odl.uniform_discr([0, 0], [1, 2], (10, 5))
    >>> discr.cell_sides
    array([ 0.1,  0.4])

    If no additional argument is given, a copy of ``discr`` is
    returned:

    >>> odl.uniform_discr_fromdiscr(discr) == discr
    True
    >>> odl.uniform_discr_fromdiscr(discr) is discr
    False

    Giving ``min_pt`` or ``max_pt`` results in a
    translation, while for the other two options, the domain
    is kept but re-partitioned:

    >>> odl.uniform_discr_fromdiscr(discr, min_pt=[1, 1])
    uniform_discr([ 1.,  1.], [ 2.,  3.], (10, 5))
    >>> odl.uniform_discr_fromdiscr(discr, max_pt=[0, 0])
    uniform_discr([-1., -2.], [ 0.,  0.], (10, 5))
    >>> odl.uniform_discr_fromdiscr(discr, cell_sides=[1, 1])
    uniform_discr([ 0.,  0.], [ 1.,  2.], (1, 2))
    >>> odl.uniform_discr_fromdiscr(discr, shape=[5, 5])
    uniform_discr([ 0.,  0.], [ 1.,  2.], (5, 5))
    >>> odl.uniform_discr_fromdiscr(discr, shape=[5, 5]).cell_sides
    array([ 0.2,  0.4])

    The cases with 2 or more additional arguments and the syntax
    for specifying quantities per axis is illustrated in the following:

    # axis 0: translate to match max_pt = 3
    # axis 1: recompute max_pt using the original shape with the
    # new min_pt and cell_sides
    >>> new_discr = odl.uniform_discr_fromdiscr(discr, min_pt=[None, 1],
    ...                                         max_pt=[3, None],
    ...                                         cell_sides=[None, 0.25])
    >>> new_discr
    uniform_discr([ 2.,  1.], [ 3.  ,  2.25], (10, 5))
    >>> new_discr.cell_sides
    array([ 0.1 ,  0.25])

    # axis 0: recompute min_pt from old cell_sides and new
    # max_pt and shape
    # axis 1: use new min_pt, shape and cell_sides only
    >>> new_discr = odl.uniform_discr_fromdiscr(discr, min_pt=[None, 1],
    ...                                         max_pt=[3, None],
    ...                                         shape=[5, 5],
    ...                                         cell_sides=[None, 0.25])
    >>> new_discr
    uniform_discr([ 2.5,  1. ], [ 3.  ,  2.25], (5, 5))
    >>> new_discr.cell_sides
    array([ 0.1 ,  0.25])
    """
    if not isinstance(discr, DiscretizedSpace):
        raise TypeError('`discr` {!r} is not a DiscretizedSpace instance'
                        ''.format(discr))
    if not discr.is_uniform:
        raise ValueError('`discr` {} is not uniformly discretized'
                         ''.format(discr))

    # Normalize partition parameters
    min_pt = normalized_scalar_param_list(min_pt, discr.ndim,
                                          param_conv=float, keep_none=True)
    max_pt = normalized_scalar_param_list(max_pt, discr.ndim,
                                          param_conv=float, keep_none=True)
    shape = normalized_scalar_param_list(shape, discr.ndim,
                                         param_conv=safe_int_conv,
                                         keep_none=True)
    cell_sides = normalized_scalar_param_list(cell_sides, discr.ndim,
                                              param_conv=float, keep_none=True)

    nodes_on_bdry = kwargs.pop('nodes_on_bdry', False)
    nodes_on_bdry = normalized_nodes_on_bdry(nodes_on_bdry, discr.ndim)

    new_min_pt = []
    new_max_pt = []
    new_shape = []
    new_csides = []
    for i, (xmin, xmax, n, s, old_xmin, old_xmax, old_n, old_s) in enumerate(
            zip(min_pt, max_pt, shape, cell_sides,
                discr.min_pt, discr.max_pt, discr.shape,
                discr.cell_sides)):
        num_params = sum(p is not None for p in (xmin, xmax, n, s))

        if num_params == 0:
            new_params = [old_xmin, old_xmax, old_n, None]

        elif num_params == 1:
            if xmin is not None:
                new_params = [xmin, old_xmax + (xmin - old_xmin), old_n, None]
            elif xmax is not None:
                new_params = [old_xmin + (xmax - old_xmax), xmax, old_n, None]
            elif n is not None:
                new_params = [old_xmin, old_xmax, n, None]
            else:
                new_params = [old_xmin, old_xmax, None, s]

        elif num_params == 2:
            if xmin is not None and xmax is not None:
                new_params = [xmin, xmax, old_n, None]
            elif xmin is not None and n is not None:
                new_params = [xmin, None, n, old_s]
            elif xmin is not None and s is not None:
                new_params = [xmin, None, old_n, s]
            elif xmax is not None and n is not None:
                new_params = [None, xmax, n, old_s]
            elif xmax is not None and s is not None:
                new_params = [None, xmax, old_n, s]
            else:
                raise ValueError('in axis {}: cannot use `shape` and '
                                 '`cell_size` only due to ambiguous values '
                                 'for `min_pt` and `max_pt`.'.format(i))

        else:
            new_params = [xmin, xmax, n, s]

        new_min_pt.append(new_params[0])
        new_max_pt.append(new_params[1])
        new_shape.append(new_params[2])
        new_csides.append(new_params[3])

    new_part = uniform_partition(min_pt=new_min_pt, max_pt=new_max_pt,
                                 shape=new_shape,
                                 cell_sides=new_csides,
                                 nodes_on_bdry=nodes_on_bdry)

    return uniform_discr_frompartition(
        new_part, exponent=discr.exponent, impl=discr.impl, **kwargs
    )


def _scaling_func_list(bdry_fracs, exponent):
    """Return a list of lists of scaling functions for the boundary."""
    def scaling(factor):
        def scaling_func(x):
            return x * factor
        return scaling_func

    func_list = []
    for frac_l, frac_r in bdry_fracs:
        func_list_entry = []
        if np.isclose(frac_l, 1.0):
            func_list_entry.append(None)
        else:
            func_list_entry.append(scaling(frac_l ** (1 / exponent)))

        if np.isclose(frac_r, 1.0):
            func_list_entry.append(None)
        else:
            func_list_entry.append(scaling(frac_r ** (1 / exponent)))

        func_list.append(func_list_entry)
    return func_list


if __name__ == '__main__':
    from odl.core.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2020 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Run all doctests in the online documentation.

Running this file causes all relevant files in the online documentation to
be run by ``doctest``, and any exception will give a FAILED result.
All other results are considered PASSED.

This test file assumes that all dependencies are installed.
"""

from __future__ import division

import doctest
import os
from doctest import ELLIPSIS, IGNORE_EXCEPTION_DETAIL, NORMALIZE_WHITESPACE

import numpy
import pytest

import odl
from odl.core.util.testutils import simple_fixture

try:
    import matplotlib
    matplotlib.use('agg')  # prevent backend from freezing
    import matplotlib.pyplot as plt
except ImportError:
    pass


doctest_extraglobs = {'odl': odl, 'np': numpy}

root_dir = os.path.join(os.path.abspath(os.path.dirname(__file__)),
                        os.pardir, os.pardir, 'doc', 'source')
root_dir = os.path.normpath(root_dir)
test_dirs = ['guide', 'getting_started']
test_suffixes = ['.rst', '.py']
exclude_files = ['faq.rst']

doc_src_files = []
doctest_optionflags = NORMALIZE_WHITESPACE | ELLIPSIS | IGNORE_EXCEPTION_DETAIL

for test_dir in test_dirs:
    for path, _, filenames in os.walk(os.path.join(root_dir, test_dir)):
        for filename in filenames:
            if (any(filename.endswith(suffix) for suffix in test_suffixes) and
                    filename not in exclude_files):
                doc_src_files.append(os.path.join(path, filename))


doc_src_file = simple_fixture("doc_src_file", doc_src_files)


@pytest.mark.suite("doc_doctests")
def test_file(doc_src_file):
    # FIXXXME: This doesn't seem to actually test the file :-(
    doctest.testfile(doc_src_file, module_relative=False, report=True,
                     extraglobs=doctest_extraglobs, verbose=True,
                     optionflags=doctest_optionflags)
    plt.close('all')


if __name__ == '__main__':
    odl.core.util.test_file(__file__)



# Copyright 2014-2017 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Test for the default functionals."""

from __future__ import division
import numpy as np
import os 
os.environ['SCIPY_ARRAY_API']='1'
import scipy.special
import pytest

import odl
from odl.core.util.testutils import all_almost_equal, noise_element, simple_fixture
from odl.functionals.default_functionals import (
    KullbackLeiblerConvexConj, KullbackLeiblerCrossEntropyConvexConj)


# --- pytest fixtures --- #


scalar = simple_fixture('scalar', [0.01, 2.7, np.array(5.0), 10, -2, -0.2,
                                   -np.array(7.1), 0])
sigma = simple_fixture('sigma', [0.001, 2.7, 10])
exponent = simple_fixture('sigma', [1, 2, 1.5, 2.5, -1.6])


space_params = ['r10', 'uniform_discr', 'power_space_unif_discr']
space_ids = [' space={} '.format(p) for p in space_params]


@pytest.fixture(scope="module", ids=space_ids, params=space_params)
def space(request, odl_impl_device_pairs):
    name = request.param.strip()
    impl, device = odl_impl_device_pairs

    if name == 'r10':
        return odl.rn(10, impl=impl, device=device)
    elif name == 'uniform_discr':
        return odl.uniform_discr(0, 1, 7, impl=impl, device=device)
    elif name == 'power_space_unif_discr':
        # Discretization parameters
        space = odl.uniform_discr(0, 1, 7, impl=impl, device=device)
        return odl.ProductSpace(space, 2)

# --- functional tests --- #


def test_L1_norm(space, sigma):
    """Test the L1-norm."""
    sigma = float(sigma)
    func = odl.functionals.L1Norm(space)
    x = noise_element(space)

    # Test functional evaluation
    expected_result = odl.abs(x).inner(space.one())
    assert func(x) == pytest.approx(expected_result)

    # Test gradient - expecting sign function
    expected_result = func.domain.element(odl.sign(x))
    assert all_almost_equal(func.gradient(x), expected_result)

    # Test proximal - expecting the following:
    #                            |  x_i + sigma, if x_i < -sigma
    #                      z_i = {  0,           if -sigma <= x_i <= sigma
    #                            |  x_i - sigma, if x_i > sigma
    tmp = space.zero().asarray()
    orig = x.asarray()
    tmp[orig > sigma] = orig[orig > sigma] - sigma
    tmp[orig < -sigma] = orig[orig < -sigma] + sigma
    expected_result = space.element(tmp)
    assert all_almost_equal(func.proximal(sigma)(x), expected_result)

    # Test convex conjugate - expecting 0 if |x|_inf <= 1, infty else
    func_cc = func.convex_conj
    norm_larger_than_one = 1.1 * x / odl.max(odl.abs(x))
    assert func_cc(norm_larger_than_one) == float('inf')

    norm_less_than_one = 0.9 * x / odl.max(odl.abs(x))
    assert func_cc(norm_less_than_one) == 0

    norm_equal_to_one = x / odl.max(odl.abs(x))
    assert func_cc(norm_equal_to_one) == 0

    # Gradient of the convex conjugate (not implemeted)
    with pytest.raises(NotImplementedError):
        func_cc.gradient

    # Test proximal of the convex conjugate - expecting x / max(1, |x|)
    expected_result = x / odl.maximum(1, odl.abs(x))
    assert all_almost_equal(func_cc.proximal(sigma)(x), expected_result)

    # Verify that the biconjugate is the functional itself
    func_cc_cc = func_cc.convex_conj
    assert isinstance(func_cc_cc, odl.functionals.L1Norm)


def test_indicator_lp_unit_ball(space, sigma, exponent):
    """Test for indicator function on unit ball."""
    x = noise_element(space)
    one_elem = space.one()

    func = odl.functionals.IndicatorLpUnitBall(space, exponent)

    # Test functional evaluation
    p_norm_x = np.power(
        func.domain.element(odl.pow(odl.abs(x), exponent)).inner(one_elem),
        1.0 / exponent)

    norm_larger_than_one = 1.01 * x / p_norm_x
    assert func(norm_larger_than_one) == np.inf

    norm_less_than_one = 0.99 * x / p_norm_x
    assert func(norm_less_than_one) == 0


def test_L2_norm(space, sigma):
    """Test the L2-norm."""
    func = odl.functionals.L2Norm(space)
    x = noise_element(space)
    x_norm = x.norm()

    # Test functional evaluation
    expected_result = np.sqrt((x ** 2).inner(space.one()))
    assert func(x) == pytest.approx(expected_result)

    # Test gradient
    if x_norm > 0:
        expected_result = x / x.norm()
        assert all_almost_equal(func.gradient(x), expected_result)

    # Verify that the gradient at zero is zero
    assert all_almost_equal(func.gradient(func.domain.zero()), space.zero())

    # Test proximal operator - expecting
    # x * (1 - sigma/||x||) if ||x|| > sigma, 0 else
    norm_less_than_sigma = 0.99 * sigma * x / x_norm
    assert all_almost_equal(func.proximal(sigma)(norm_less_than_sigma),
                            space.zero())

    norm_larger_than_sigma = 1.01 * sigma * x / x_norm
    expected_result = (norm_larger_than_sigma *
                       (1.0 - sigma / norm_larger_than_sigma.norm()))
    assert all_almost_equal(func.proximal(sigma)(norm_larger_than_sigma),
                            expected_result)

    # Test convex conjugate
    func_cc = func.convex_conj

    # Test evaluation of the convex conjugate - expecting
    # 0 if ||x|| < 1, infty else
    norm_larger_than_one = 1.01 * x / x_norm
    assert func_cc(norm_larger_than_one) == np.inf

    norm_less_than_one = 0.99 * x / x_norm
    assert func_cc(norm_less_than_one) == 0

    # Gradient of the convex conjugate (not implemeted)
    with pytest.raises(NotImplementedError):
        func_cc.gradient

    # Test the proximal of the convex conjugate - expecting
    # x if ||x||_2 < 1, x/||x|| else
    if x_norm < 1:
        expected_result = x
    else:
        expected_result = x / x_norm
    assert all_almost_equal(func_cc.proximal(sigma)(x), expected_result)

    # Verify that the biconjugate is the functional itself
    func_cc_cc = func_cc.convex_conj
    assert func_cc_cc(x) == pytest.approx(func(x))


def test_L2_norm_squared(space, sigma):
    """Test the squared L2-norm."""
    func = odl.functionals.L2NormSquared(space)
    x = noise_element(space)
    x_norm = x.norm()

    # Test functional evaluation
    expected_result = x_norm ** 2
    assert func(x) == pytest.approx(expected_result)

    # Test gradient
    expected_result = 2.0 * x
    assert all_almost_equal(func.gradient(x), expected_result)

    # Test proximal operator
    expected_result = x / (1 + 2.0 * sigma)
    assert all_almost_equal(func.proximal(sigma)(x), expected_result)

    # Test convex conjugate
    func_cc = func.convex_conj

    # Test evaluation of the convex conjugate
    expected_result = x_norm ** 2 / 4.0
    assert func_cc(x) == pytest.approx(expected_result)

    # Test gradient of the convex conjugate
    expected_result = x / 2.0
    assert all_almost_equal(func_cc.gradient(x), expected_result)

    # Test proximal of the convex conjugate
    expected_result = x / (1 + sigma / 2.0)
    assert all_almost_equal(func_cc.proximal(sigma)(x), expected_result)

    # Verify that the biconjugate is the functional itself
    func_cc_cc = func_cc.convex_conj

    # Check that they evaluate to the same value
    assert func_cc_cc(x) == pytest.approx(func(x))

    # Check that their gradients evaluate to the same value
    assert all_almost_equal(func_cc_cc.gradient(x), func.gradient(x))


def test_constant_functional(space, scalar):
    """Test the constant functional."""
    constant = float(scalar)
    func = odl.functionals.ConstantFunctional(space, constant=scalar)
    x = noise_element(space)

    assert func.constant == constant

    # Test functional evaluation
    assert func(x) == constant

    # Test gradient - expecting zero operator
    assert isinstance(func.gradient, odl.ZeroOperator)

    # Test proximal operator - expecting identity
    sigma = 1.5
    assert isinstance(func.proximal(sigma), odl.IdentityOperator)

    # Test convex conjugate
    func_cc = func.convex_conj

    # Test evaluation of the convex conjugate - expecting
    # -constant if x=0, infty else
    assert func_cc(x) == np.inf
    assert func_cc(space.zero()) == -constant

    # Gradient of the convex conjugate (not implemeted)
    with pytest.raises(NotImplementedError):
        func_cc.gradient

    # Proximal of the convex conjugate - expecting zero operator
    assert isinstance(func_cc.proximal(sigma), odl.ZeroOperator)

    # Verify that the biconjugate is the functional itself
    func_cc_cc = func_cc.convex_conj
    assert isinstance(func_cc_cc, odl.functionals.ConstantFunctional)
    assert func_cc_cc.constant == constant


def test_zero_functional(space):
    """Test the zero functional."""
    zero_func = odl.functionals.ZeroFunctional(space)
    assert isinstance(zero_func, odl.functionals.ConstantFunctional)
    assert zero_func.constant == 0


def test_kullback_leibler(space):
    """Test the kullback leibler functional and its convex conjugate."""
    # The prior needs to be positive
    prior = odl.abs(noise_element(space)) + 0.1

    func = odl.functionals.KullbackLeibler(space, prior)

    # The fucntional is only defined for positive elements
    x = odl.abs(noise_element(space)) + 0.1
    one_elem = space.one()

    # Evaluation of the functional
    expected_result = (
        x - prior + prior * odl.log(prior / x)
    ).inner(one_elem)
    assert func(x) == pytest.approx(expected_result)

    # Check property for prior
    assert all_almost_equal(func.prior, prior)

    # For elements with (a) negative components it should return inf
    x_neg = noise_element(space)
    x_neg = x_neg - odl.max(x_neg)
    assert func(x_neg) == np.inf

    # The gradient
    expected_result = 1 - prior / x
    assert all_almost_equal(func.gradient(x), expected_result)

    # The proximal operator
    sigma = np.random.rand()
    expected_result = odl.solvers.proximal_convex_conj(
        odl.solvers.proximal_convex_conj_kl(space, g=prior))(sigma)(x)
    assert all_almost_equal(func.proximal(sigma)(x), expected_result)

    # The convex conjugate functional
    cc_func = func.convex_conj

    assert isinstance(cc_func, KullbackLeiblerConvexConj)

    # The convex conjugate functional is only finite for elements with all
    # components smaller than 1.
    x = noise_element(space)
    x = x - odl.max(x) + 0.99

    # Evaluation of convex conjugate
    expected_result = - (prior * odl.log(1 - x)).inner(one_elem)
    assert cc_func(x) == pytest.approx(expected_result)

    x_wrong = noise_element(space)
    x_wrong = x_wrong - odl.max(x_wrong) + 1.01
    assert cc_func(x_wrong) == np.inf

    # The gradient of the convex conjugate
    expected_result = prior / (1 - x)
    assert all_almost_equal(cc_func.gradient(x), expected_result)

    # The proximal of the convex conjugate
    expected_result = 0.5 * (1 + x - odl.sqrt((x - 1) ** 2 + 4 * sigma * prior))
    assert all_almost_equal(cc_func.proximal(sigma)(x), expected_result)

    # The biconjugate, which is the functional itself since it is proper,
    # convex and lower-semicontinuous
    cc_cc_func = cc_func.convex_conj

    # Check that they evaluate the same
    assert cc_cc_func(x) == pytest.approx(func(x))


def test_kullback_leibler_cross_entropy(space):
    """Test the kullback leibler cross entropy and its convex conjugate."""
    # The prior needs to be positive
    prior = noise_element(space)
    prior = space.element(odl.abs(prior))

    func = odl.functionals.KullbackLeiblerCrossEntropy(space, prior)

    # The fucntional is only defined for positive elements
    x = noise_element(space)
    x = func.domain.element(odl.abs(x))
    one_elem = space.one()

    # Evaluation of the functional
    expected_result = ((prior - x + x * odl.log(x / prior))
                       .inner(one_elem))
    assert func(x) == pytest.approx(expected_result)

    # Check property for prior
    assert all_almost_equal(func.prior, prior)

    # For elements with (a) negative components it should return inf
    x_neg = noise_element(space)
    x_neg = x_neg - odl.max(x_neg)
    assert func(x_neg) == np.inf

    # The gradient
    expected_result = odl.log(x / prior)
    assert all_almost_equal(func.gradient(x), expected_result)

    # The proximal operator
    sigma = np.random.rand()
    prox = odl.solvers.proximal_convex_conj(
        odl.solvers.proximal_convex_conj_kl_cross_entropy(space, g=prior))
    expected_result = prox(sigma)(x)
    assert all_almost_equal(func.proximal(sigma)(x), expected_result)

    # The convex conjugate functional
    cc_func = func.convex_conj

    assert isinstance(cc_func, KullbackLeiblerCrossEntropyConvexConj)

    # The convex conjugate functional is defined for all values of x.
    x = noise_element(space)

    # Evaluation of convex conjugate
    expected_result = (prior * (odl.exp(x) - 1)).inner(one_elem)
    assert cc_func(x) == pytest.approx(expected_result)

    # The gradient of the convex conjugate
    expected_result = prior * odl.exp(x)
    assert all_almost_equal(cc_func.gradient(x), expected_result)

    # The proximal of the convex conjugate
    if isinstance(space, odl.ProductSpace):
        device  = space[0].device
        backend = space[0].array_backend
    else:
        device = space.device
        backend = space.array_backend
    arr = backend.to_cpu((prior * odl.exp(x)).asarray())
    x_arr = backend.to_cpu(x.asarray())

    
    expected_result = x_arr - scipy.special.lambertw(sigma * arr).real
    if device != 'cpu':
        expected_result = expected_result.to(device)
    if not all_almost_equal(cc_func.proximal(sigma)(x), expected_result):
        print(f'{cc_func.proximal(sigma)(x)=}')
        print(f'{expected_result=}')
        assert False

    # The biconjugate, which is the functional itself since it is proper,
    # convex and lower-semicontinuous
    cc_cc_func = cc_func.convex_conj

    # Check that they evaluate the same
    assert cc_cc_func(x) == pytest.approx(func(x))


def test_quadratic_form(space):
    """Test the quadratic form functional."""
    operator = odl.IdentityOperator(space)
    vector = space.one()
    constant = 0.363
    func = odl.functionals.QuadraticForm(operator, vector, constant)

    x = noise_element(space)

    # Checking that values is stored correctly
    assert func.operator == operator
    assert func.vector == vector
    assert func.constant == constant

    # Evaluation of the functional
    expected_result = x.inner(operator(x)) + vector.inner(x) + constant
    assert func(x) == pytest.approx(expected_result)

    # The gradient
    expected_gradient = 2 * operator(x) + vector
    assert all_almost_equal(func.gradient(x), expected_gradient)

    # The convex conjugate
    assert isinstance(func.convex_conj, odl.functionals.QuadraticForm)

    # Test for linear functional
    func_no_operator = odl.functionals.QuadraticForm(vector=vector,
                                                 constant=constant)
    expected_result = vector.inner(x) + constant
    assert func_no_operator(x) == pytest.approx(expected_result)

    expected_gradient = vector
    assert all_almost_equal(func_no_operator.gradient(x), expected_gradient)

    # The convex conjugate is a translation of the IndicatorZero
    func_no_operator_cc = func_no_operator.convex_conj
    assert isinstance(func_no_operator_cc,
                      odl.functionals.FunctionalTranslation)
    assert isinstance(func_no_operator_cc.functional,
                      odl.functionals.IndicatorZero)
    assert func_no_operator_cc(vector) == -constant
    assert odl.isinf(func_no_operator_cc(vector + 2.463))

    # Test with no offset
    func_no_offset = odl.functionals.QuadraticForm(operator, constant=constant)
    expected_result = x.inner(operator(x)) + constant
    assert func_no_offset(x) == pytest.approx(expected_result)


def test_separable_sum(space):
    """Test for the separable sum."""
    l1 = odl.functionals.L1Norm(space)
    l2 = odl.functionals.L2Norm(space)

    x = noise_element(space)
    y = noise_element(space)

    # Initialization and calling
    func = odl.functionals.SeparableSum(l1, l2)
    assert func([x, y]) == pytest.approx(l1(x) + l2(y))

    power_func = odl.functionals.SeparableSum(l1, 5)
    assert power_func([x, x, x, x, x]) == pytest.approx(5 * l1(x))

    # Gradient
    grad = func.gradient([x, y])
    assert grad[0] == l1.gradient(x)
    assert grad[1] == l2.gradient(y)

    # Proximal
    sigma = 1.0
    prox = func.proximal(sigma)([x, y])
    assert prox[0] == l1.proximal(sigma)(x)
    assert prox[1] == l2.proximal(sigma)(y)

    # Convex conjugate
    assert func.convex_conj([x, y]) == l1.convex_conj(x) + l2.convex_conj(y)


def test_moreau_envelope_l1():
    """Test for the Moreau envelope with L1 norm."""
    space = odl.rn(3)

    l1 = odl.functionals.L1Norm(space)

    # Test l1 norm, gives "Huber norm"
    smoothed_l1 = odl.functionals.MoreauEnvelope(l1)
    assert all_almost_equal(smoothed_l1.gradient([0, -0.2, 0.7]),
                            [0, -0.2, 0.7])
    assert all_almost_equal(smoothed_l1.gradient([-3, 2, 10]),
                            [-1, 1, 1])

    # Test with different sigma
    smoothed_l1 = odl.functionals.MoreauEnvelope(l1, sigma=0.5)
    assert all_almost_equal(smoothed_l1.gradient([0, 0.2, 0.7]),
                            [0, 0.4, 1.0])


def test_moreau_envelope_l2_sq(space, sigma):
    """Test for the Moreau envelope with l2 norm squared."""

    # Result is ||x||_2^2 / (1 + 2 sigma)
    # Gradient is x * 2 / (1 + 2 * sigma)
    l2_sq = odl.functionals.L2NormSquared(space)

    smoothed_l2_sq = odl.functionals.MoreauEnvelope(l2_sq, sigma=sigma)
    x = noise_element(space)
    assert all_almost_equal(smoothed_l2_sq.gradient(x),
                            x * 2 / (1 + 2 * sigma))


def test_weighted_separablesum(space):
    """Test for the weighted proximal of a SeparableSum functional."""

    l1 = odl.functionals.L1Norm(space)
    l2 = odl.functionals.L2Norm(space)
    func = odl.functionals.SeparableSum(l1, l2)

    x = func.domain.one()

    sigma = [0.5, 1.0]

    prox = func.proximal(sigma)(x)
    assert all_almost_equal(prox, [l1.proximal(sigma[0])(x[0]),
                                   l2.proximal(sigma[1])(x[1])])


def test_weighted_proximal_L2_norm_squared(space):
    """Test for the weighted proximal of the squared L2 norm"""

    # Define the functional on the space.
    func = odl.functionals.L2NormSquared(space)

    # Set the stepsize as a random element of the spaces
    # with elements between 1 and 10.
    sigma = odl.core.phantom.uniform_noise(space, 1, 10)

    # Start at the one vector.
    x = space.one()

    # Calculate the proximal point in-place and out-of-place
    p_ip = space.element()
    func.proximal(sigma)(x, out=p_ip)
    p_oop = func.proximal(sigma)(x)

    # Both should contain the same vector now.
    assert all_almost_equal(p_ip, p_oop)

    # Check if the subdifferential inequalities are satisfied.
    # p = prox_{sigma * f}(x) iff (x - p)/sigma = grad f(p)
    assert all_almost_equal(func.gradient(p_ip),
                            (x - p_ip) / sigma)


def test_weighted_proximal_L1_norm_far(space):
    """Test for the weighted proximal of the L1 norm away from zero"""

    # Define the functional on the space.
    func = odl.functionals.L1Norm(space)

    # Set the stepsize as a random element of the spaces
    # with elements between 1 and 10.
    sigma = odl.core.phantom.noise.uniform_noise(space, 1, 10)

    # Start far away from zero so that the L1 norm will be differentiable
    # at the result.
    x = 100 * space.one()

    # Calculate the proximal point in-place and out-of-place
    p_ip = space.element()
    func.proximal(sigma)(x, out=p_ip)
    p_oop = func.proximal(sigma)(x)

    # Both should contain the same vector now.
    assert all_almost_equal(p_ip, p_oop)

    # Check if the subdifferential inequalities are satisfied.
    # p = prox_{sigma * f}(x) iff (x - p)/sigma = grad f(p)
    assert all_almost_equal(func.gradient(p_ip), (x - p_ip) / sigma)


def test_weighted_proximal_L1_norm_close(space):
    """Test for the weighted proximal of the L1 norm near zero"""

    # Set the space.
    space = odl.rn(5)

    # Define the functional on the space.
    func = odl.functionals.L1Norm(space)

    # Set the stepsize.
    sigma = [0.1, 0.2, 0.5, 1.0, 2.0]

    # Set the starting point.
    x = 0.5 * space.one()

    # Calculate the proximal point in-place and out-of-place
    p_ip = space.element()
    func.proximal(sigma)(x, out=p_ip)
    p_oop = func.proximal(sigma)(x)

    # Both should contain the same vector now.
    assert all_almost_equal(p_ip, p_oop)

    # Check if this equals the expected result.
    expected_result = [0.4, 0.3, 0.0, 0.0, 0.0]
    assert all_almost_equal(expected_result, p_ip)


def test_bregman_functional_no_gradient(space):
    """Test Bregman distance for functional without gradient."""

    ind_func = odl.functionals.IndicatorNonnegativity(space)
    point = odl.abs(noise_element(space))
    subgrad = noise_element(space)  # Any element in the domain is ok
    bregman_dist = odl.functionals.BregmanDistance(ind_func, point, subgrad)

    x = odl.abs(noise_element(space))

    expected_result = -subgrad.inner(x - point)
    assert all_almost_equal(bregman_dist(x), expected_result)

    # However, since the functional is not differentialbe we cannot call the
    # gradient of the Bregman distance functional
    with pytest.raises(NotImplementedError):
        bregman_dist.gradient


def test_bregman_functional_l2_squared(space, sigma):
    """Test Bregman distance using l2 norm squared as underlying functional."""
    sigma = float(sigma)

    l2_sq = odl.functionals.L2NormSquared(space)
    point = noise_element(space)
    subgrad = l2_sq.gradient(point)
    bregman_dist = odl.functionals.BregmanDistance(l2_sq, point, subgrad)

    expected_func = odl.functionals.L2NormSquared(space).translated(point)

    x = noise_element(space)

    # Function evaluation
    assert all_almost_equal(bregman_dist(x), expected_func(x))

    # Gradient evaluation
    assert all_almost_equal(bregman_dist.gradient(x),
                            expected_func.gradient(x))

    # Convex conjugate
    cc_bregman_dist = bregman_dist.convex_conj
    cc_expected_func = expected_func.convex_conj
    assert all_almost_equal(cc_bregman_dist(x), cc_expected_func(x))

    # Proximal operator
    prox_bregman_dist = bregman_dist.proximal(sigma)
    prox_expected_func = expected_func.proximal(sigma)
    assert all_almost_equal(prox_bregman_dist(x), prox_expected_func(x))


if __name__ == '__main__':
    odl.core.util.test_file(__file__)
